# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.

from django.utils.translation import ugettext as _
from basesoft.web.contrib.contact.models import Company
from basesoft.web.contrib.license.models import License
from basesoft.web.contrib.order.models import Currency
from basesoft.web.core.util.custom_query import CustomQueryModel, CustomQueryField
from sqlalchemy.sql.expression import case
from sqlalchemy.sql.functions import func
from sqlalchemy.orm.util import aliased
from we.core.custom_query import AbstractWeCustomQuery
from we.declaration.models import WeDeclaration, WeDeclarationLine
from we.member.models import WeLpp, WeHqppLicense
from we.product.models import WeHqppProduct, WeHqppRegulationPeriodProduct
from we.tbs.models import WeTbsLicense

ArCompany = aliased(Company)
InvoiceCompany = aliased(Company)
Tbs = aliased(License)

class WeDeclarationCustomQuery(AbstractWeCustomQuery):
    permissions = ('we_view_declaration', )

    def tables(self):
        return (
            CustomQueryModel(WeDeclaration),
            CustomQueryModel(WeLpp, onclause=WeDeclaration.lpp_id == WeLpp.id),
            CustomQueryModel(WeDeclarationLine, onclause=WeDeclarationLine.declaration_id == WeDeclaration.id, use_outerjoin=True),
            CustomQueryModel(WeHqppRegulationPeriodProduct,
                             onclause=WeDeclarationLine.hqpp_period_product_id == WeHqppRegulationPeriodProduct.id, use_outerjoin=True),
            CustomQueryModel(WeHqppProduct, onclause=WeHqppRegulationPeriodProduct.product_id == WeHqppProduct.id, use_outerjoin=True),
            CustomQueryModel(Currency, onclause=WeDeclaration.currency_id == Currency.id, use_outerjoin=True),
            CustomQueryModel(WeHqppLicense, onclause=WeLpp.hqpp_id == WeHqppLicense.license_id, use_outerjoin=True),
            CustomQueryModel(Tbs, onclause=WeLpp.tbs_id == Tbs.id, use_outerjoin=True),
            CustomQueryModel(ArCompany, onclause=WeLpp.ar_company_id == ArCompany.id, use_outerjoin=True),
            CustomQueryModel(InvoiceCompany, onclause=WeLpp.invoice_company_id == InvoiceCompany.id, use_outerjoin=True),
        )

    def fields(self):
        return (
            CustomQueryField(WeDeclaration.id, _('Declaration ID'), key='declaration_id', is_filter=True),
            CustomQueryField(WeDeclaration.lpp_id, _('LPP ID'), key='lpp_id'),
            CustomQueryField(WeLpp.name, _('LPP'), key='lpp_name', is_filter=True),
            CustomQueryField(WeHqppLicense.name, _('Producer'), key='hqpp_name', is_filter=True),
            CustomQueryField(Tbs.name, _('TBS'), key='tbs_name', is_filter=True),
            CustomQueryField(WeDeclaration.status, _('Status'), key='status', is_filter=True),
            CustomQueryField(func.group_concat(" ", WeHqppProduct.sku), _('SKU'), key='product_sku', is_filter=True),
            CustomQueryField(WeDeclaration.year, _('Year'), key='year', is_filter=True),
            CustomQueryField(WeDeclaration.month, _('Month'), key='month', is_filter=True),
            CustomQueryField(WeDeclaration.total_weight, _('Weight'), key='total_weight', use_when_separated=WeDeclarationLine.product_total_weight),
            CustomQueryField(WeDeclaration.total_quantity, _('Quantity'), key='total_quantity', use_when_separated=WeDeclarationLine.quantity),
            CustomQueryField(case([(Currency.symbol == None, Currency.code)], else_=Currency.symbol), _('Currency'), key='currency_symbol'),
            CustomQueryField(WeDeclaration.total_fee, _('Amount'), key='total_fee', use_when_separated=WeDeclarationLine.product_total_fee),
            CustomQueryField(ArCompany.name, _('Authorized representative'), key='ar_company_name', is_filter=True),
            CustomQueryField(InvoiceCompany.name, _('Invoice company'), key='invoice_company_name', is_filter=True),
        )

    def set_where(self):
        acc_license = self.get_default_group().license
        if isinstance(acc_license, WeTbsLicense):
            self.where = WeLpp.id.in_([x.id for x in self.get_local_producer_partners(tbs_id=acc_license.id)])
        if isinstance(acc_license, WeHqppLicense):
            self.where = WeLpp.hqpp_id.in_([x.id for x in self.get_member_licenses()])

    def set_group_by(self):
        if self.separated:
            self.group_by = (WeDeclaration.id, WeHqppProduct.sku, )
        else:
            self.group_by = (WeDeclaration.id, )

custom_queries = (
    WeDeclarationCustomQuery,
)