# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
import os

from datetime import date
from django.utils.translation import ugettext as _, ugettext_lazy
from basesoft.web.contrib.file.controllers import ImportFileController
from basesoft.web.contrib.notification.controllers import NotificationController
from basesoft.web.core.exceptions import PublicSecurityException, PublicUserException
from basesoft.web.core.importers import Utf8CsvImporter, XlsxImporter
from sqlalchemy.sql.expression import and_
from basesoft.web.core.util.controllers import transaction_decorator
from we.core.exceptions import SKUNotAvailableException
from we.core.importers import WeImportException
from we.declaration.controllers import WeDeclarationController
from we.declaration.models import WeDeclaration
from we.member.models import WeLpp, WeHqppLicense
from we.product.controllers import WeProductController
from we.product.models import WeHqppRegulationPeriodProduct, WeHqppProduct, WeHqppProductExcludedAtTbs
from we.tbs.models import WeTbsRegulationPeriod


class WeDeclarationAbstractImporter(ImportFileController):
    IMPORT_FILE_TYPE = 'we.declaration.Declaration'
    IMPORT_FILE_TYPE_I18N = ugettext_lazy('Declaration')

    @staticmethod
    def get_importer(tmp_file):
        """
        static method to return the right importer for the given file
        """
        base_filename, file_extension = os.path.splitext(tmp_file.file_name)

        importer = None
        if file_extension in ('.csv', '.txt'):
            importer = WeDeclarationCsvImporter

        elif file_extension == '.xlsx':
            importer = WeDeclarationXlsxImporter

        else:
            raise Exception(_('Cannot import for extension: %s') % file_extension)

        return importer

    def get_human_readable_settings(self, import_settings):
        rv = {}

        dc = WeDeclarationController(self.sa_session, account_id=self.get_account().id)
        hqpp = dc.get_member_licenses(import_settings['hqpp_id']).pop()

        rv[_('HQPP')] = hqpp.name
        return rv


INDEX_LPP = 1
INDEX_YEAR = 2
INDEX_MONTH = 3
INDEX_SKU = 4
INDEX_B2B = 5
INDEX_QUANTITY = 6

class WeDeclarationImporter(WeDeclarationAbstractImporter):

    MAX_RECORDS_FOR_INSTANT_IMPORT = 0
    NUMBER_EXPECTED_COLUMNS = 7
    STOP_AFTER_NUMBER_EMPTY_ROWS = 3

    import_state = None
    callback = None

    def __init__(self, session, **kwargs):
        ImportFileController.__init__(self, session, **kwargs)
        self.import_state = {}
        self._missing_data = []
        self.row_number = 0

    @staticmethod
    def get_notification_callback():
        return {
            'js_module': 'we_declaration_manager'
        }

    @staticmethod
    def get_failure_notification_callback():
        return {
            'js_module': 'we_declaration_import_wizard'
        }

    @staticmethod
    def get_notification_subject():
        return None

    @staticmethod
    def get_notification_text():
        return None

    @staticmethod
    def get_notification_update_notify():
        return ['we.product.FileImportCustomQuery.newData', 'we.declaration.WeDeclarationCustomQuery.newData']

    @transaction_decorator
    def _notify(self, file, callback, subject, text, notify):
        if len(self._missing_data) == 0:
            super(WeDeclarationImporter, self)._notify(file, callback, subject, text, notify)
            return

        notification = None
        nc = NotificationController(self.sa_session)

        notification = self._notify_failure(file, nc, {
                'js_module': 'we_declaration_missing_sku_panel',
                'parameters': [self._missing_data]
            }, 'Import failed due to missing SKU data', 'Import failed due to missing SKU data', notify)

        nc.create_system_notification(notification, file.groups[0].group)

    def _import_row(self, row, controller, dryrun=False, hqpp_id=None, row_count=None, **kwargs):
        self.row_number += 1

        if row[6] in [None, '']:
            return
        try:
            lpp_id = int(row[INDEX_LPP])
            year = int(row[INDEX_YEAR])
            month = int(row[INDEX_MONTH])
            sku = unicode(row[INDEX_SKU])
            is_b2b = row[INDEX_B2B] if row[INDEX_B2B] not in [None, ''] else None
            quantity = int(row[INDEX_QUANTITY])
        except ValueError:
            raise WeImportException(_('Could not convert data to an integer.'))

        hqpp = self.sa_session.query(WeHqppLicense).filter(WeHqppLicense.license_id == hqpp_id).first()
        if hqpp is None:
            raise WeImportException(_('There is no HQPP with ID %s.' % hqpp_id))

        lpp = self.sa_session.query(WeLpp).filter(WeLpp.id == lpp_id, WeLpp.hqpp_id == hqpp_id).first()
        if lpp is None: # or unicode(lpp_id.strip() == ''):
            raise WeImportException(_("Producer %(producer)s has no access to LPP with ID %(id)s " % {'producer': hqpp.name, 'id': lpp_id}))
        elif hqpp != lpp.hqpp:
            raise PublicSecurityException(_('The provided local producer partner is not part of the selected HQPP'))

        if is_b2b in [None, '']:
            is_b2b = WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE
        elif is_b2b.lower() == 'b2b':
            is_b2b = WeHqppRegulationPeriodProduct.IS_B2B_TRUE
        else:
            is_b2b = WeHqppRegulationPeriodProduct.IS_B2B_FALSE

        if is_b2b == WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE:
            if lpp.tbs.get_config('we.tbs.require.end.user.distinction') is True:
                raise WeImportException(_("Line %(line)s with SKU %(sku)s "
                                          "requires B2B/B2C distinction." % {'line': self.row_number, 'sku': sku}))

            if lpp.tbs.get_config('we.tbs.require.end.user.distinction') is True:
                self._missing_data.append([self.row_number,
                                           {'sku': sku},
                                           _("Line %(line)s with SKU %(sku)s "
                                       "requires B2B/B2C distinction." % {'line': self.row_number, 'sku': sku})])
                raise WeImportException(_("Line %(line)s contains missing SKU's" % {'line': self.row_number}))

        product = self.sa_session.query(WeHqppProduct).filter(
            and_(
                WeHqppProduct.hqpp == hqpp,
                WeHqppProduct.sku == sku
            )
        ).first()
        if product is None and sku not in self._missing_data:
            self._missing_data.append([self.row_number,
                                       {'sku': sku},_('Missing product for SKU.')])
            raise WeImportException(_("Line %(line)s contains missing data" % {'line': self.row_number}))

        wpc = WeProductController(self.sa_session)
        period = date(year, month, 1)
        regulation_period = wpc.get_regulation_periods(lpp.tbs, start_date=period)
        if regulation_period is None:
            self._missing_data.append([self.row_number, {
                'sku': sku}, _('No regulation period found for year %(year)d and month %(month)d at take-back scheme %(tbs)s' % {
                'year': year, 'month': month, 'tbs': lpp.tbs.name
            })])
            raise WeImportException(_("Line %(line)s contains missing SKU's" % {'line': self.row_number}))

        product_tbs_category = self.sa_session.query(WeHqppRegulationPeriodProduct).filter(
            and_(
                WeHqppRegulationPeriodProduct.product == product,
                WeHqppRegulationPeriodProduct.tbs_regulation_period == regulation_period,
                WeHqppRegulationPeriodProduct.is_b2b == is_b2b
            )
        ).first()

        excluded = self.sa_session.query(WeHqppProductExcludedAtTbs).filter(
            and_(
                WeHqppProductExcludedAtTbs.product == product,
                WeHqppProductExcludedAtTbs.tbs == lpp.tbs,
                WeHqppProductExcludedAtTbs.is_b2b == is_b2b,
            )
        ).first()

        # This product is excluded, we can skip it.
        if excluded is not None:
            return True, self._missing_data

        if product is not None and product_tbs_category is None:
            self._missing_data.append([self.row_number,
                                       {'sku': sku},
                                       _('Missing local category '
                                   'for regulation period %(period)s '
                                   'at take-back scheme %(tbs)s.') % {'period': regulation_period.title,'tbs': lpp.tbs.name}])
            raise WeImportException(_("Line %(line)s contains missing SKU's" % {'line': self.row_number}))

        # FIXME: check below is not unit tested
        if product_tbs_category.tbs_regulation_period_product.fee is None:
            self._missing_data.append([self.row_number, {
                'sku': sku}, _(
                'No fee set for product %(product)s for year %(year)d and month %(month)d at take-back scheme %(tbs)s' % {
                    'product': product_tbs_category.tbs_regulation_period_product.tbs_product_level.label,
                    'year': year, 'month': month, 'tbs': lpp.tbs.name
                })])
            raise WeImportException(
                _("Line %(line)s contains a product which has no fee set" % {'line': self.row_number}))

        if not self.import_state.has_key(lpp_id):
            self.import_state[lpp_id] = {}
        if not self.import_state[lpp_id].has_key(year):
            self.import_state[lpp_id][year] = {}
        if not self.import_state[lpp_id][year].has_key(month):
            self.import_state[lpp_id][year][month] = {}

        # add line to the state
        self.import_state[lpp_id][year][month]['lpp_id'] = lpp_id
        self.import_state[lpp_id][year][month]['year'] = year
        self.import_state[lpp_id][year][month]['month'] = month
        if not self.import_state[lpp_id][year][month].has_key('lines'):
            self.import_state[lpp_id][year][month]['lines'] = []

        line = {
            'line_id': None,
            'sku': sku,
            'quantity': quantity,
            'is_b2b': is_b2b
        }

        for l in self.import_state[lpp_id][year][month]['lines']:

            if l['sku'] == sku and l['is_b2b'] == is_b2b:
                raise WeImportException(_("Line %(line)s contains duplicate elements." % {
                                              'line': self.row_number}))

        self.import_state[lpp_id][year][month]['lines'].append(line)

        return True, self._missing_data

    def _post_read_rows(self, controller, **kwargs):
        """
        called after above all _import_row() is done
        """

        if len(self._missing_data) > 0:
            raise SKUNotAvailableException("The file contains SKUs which are not available in the Product Portfolio")

        for lpp_id in self.import_state.keys():
            for year in self.import_state[lpp_id].keys():
                for month in self.import_state[lpp_id][year].keys():
                    loaded_state = {}  # reset loaded state, as it is a by Declaration month state (#500)
                    _id = None

                    lpp = self.sa_session.query(WeLpp).filter(WeLpp.id == lpp_id).first()

                    declaration = self.sa_session.query(WeDeclaration)\
                        .filter(and_(WeDeclaration.lpp == lpp,
                                     WeDeclaration.year == year,
                                     WeDeclaration.month == month)).first()

                    if declaration is not None:
                        loaded_state = controller.load_declaration(declaration.id)
                        _id = declaration.id

                    # check for regulation period
                    _filter = and_(WeTbsRegulationPeriod.tbs == lpp.tbs,
                                   WeTbsRegulationPeriod.start_date <= date(year, month, 1),
                                   WeTbsRegulationPeriod.archived == 0)
                    regulation_period = self.sa_session.query(WeTbsRegulationPeriod).filter(_filter)\
                        .order_by(WeTbsRegulationPeriod.start_date.desc()).first()

                    if regulation_period is None:
                        raise WeImportException(_("Cannot import declaration. There are no regulation "
                                                  "periods for take-back scheme %s." % lpp.tbs.name))

                    # create a state from old lines and new lines
                    lines_state = self._create_declaration_state(lpp.hqpp, regulation_period,
                                                                 self.import_state[lpp_id][year][month]['lines'],
                                                                 loaded_state)

                    self.import_state[lpp_id][year][month]['lines'] = lines_state
                    controller.save_or_update_declaration(_id, self.import_state[lpp_id][year][month])

    def _create_declaration_state(self, hqpp, regulation_period, lines, loaded_state):

        for line in lines:
            matched = False

            # if sku is in loaded state lines set the line id and the hqpp period product
            # else add new line
            if loaded_state.has_key('lines'):
                for l in loaded_state['lines']:
                    if line['sku'].lower() == l['sku'].lower():

                        if line['is_b2b'] != l['is_b2b']:
                            continue

                        line['hqpp_period_product_id'] = l['hqpp_period_product_id']
                        line['line_id'] = l['line_id']
                        matched = True

            if matched is False:
                # get the hqpp product corresponding to the SKU
                product = self.sa_session.query(WeHqppProduct).filter(
                    WeHqppProduct.hqpp == hqpp,
                    WeHqppProduct.sku == line['sku']
                ).one()

                product_tbs_category = self.sa_session.query(WeHqppRegulationPeriodProduct).filter(
                    and_(
                        WeHqppRegulationPeriodProduct.product == product,
                        WeHqppRegulationPeriodProduct.tbs_regulation_period == regulation_period,
                        WeHqppRegulationPeriodProduct.is_b2b == line['is_b2b']
                    )
                ).first()
                line['hqpp_period_product_id'] = product_tbs_category.id
        return lines

class WeDeclarationCsvImporter(WeDeclarationImporter, Utf8CsvImporter):
    """
    Importer for CSV
    """

    def read_row(self, file_path):
        return self._read_row(file_path)


class WeDeclarationXlsxImporter(WeDeclarationImporter, XlsxImporter):
    """
    Importer for XLSX
    """

    def read_row(self, file_path):
        # return self._read_row(file_path)

        for row in self._read_row(file_path):
            values = []
            cell_i = 0
            for value in row:
                cell_i += 1
                if cell_i > 7:
                    break

                if type(value) == float:
                    value = int(value)

                values.append(value)

            yield values


import_file_map = {
    'we.declaration.Declaration': WeDeclarationAbstractImporter,
}
