# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.

import datetime

from sqlalchemy import Integer, Column, UnicodeText, DateTime, SmallInteger, ForeignKey
from sqlalchemy.orm import mapper, relation
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy.types import Unicode

from basesoft.web.contrib.license import models as license_models
from basesoft.sqlalchemy.table import IQTable

from we.member import models as member_models

__mapping_defined = False
__tables_defined = False
tables = {}

def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False

def get_tables():
    return tables

def do_tables(metadata):
    """
    This function is called by Django middleware when needed.
    """
    global __tables_defined
    if __tables_defined:
        return

    tblWeHqppLicense = IQTable('tblWeHqppLicense', metadata,
        Column('license_id', Integer, ForeignKey('tblLicense.id', ondelete='CASCADE'), primary_key=True)
    )

    tblWeLpp = IQTable('tblWeLpp', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('hqpp_id', Integer, ForeignKey('tblWeHqppLicense.license_id', ondelete='CASCADE'), nullable=False),
        Column('tbs_id', Integer, ForeignKey('tblWeTbsLicense.license_id', ondelete='CASCADE'), nullable=False),
        Column('name', UnicodeText, nullable=False),

        Column('weee_registration_number', Unicode(255)),

        Column('company_id', Integer, ForeignKey('tblCompany.id', ondelete='RESTRICT'), nullable=False),
        Column('ar_company_id', Integer, ForeignKey('tblCompany.id', ondelete='RESTRICT')),
        Column('invoice_company_id', Integer, ForeignKey('tblCompany.id', ondelete='RESTRICT')),
        Column('invoice_relation_id', UnicodeText),
        Column('lpp_relation_id', UnicodeText),

        Column('status', SmallInteger, nullable=False, default=member_models.WeLpp.STATUS_NEW),
        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),

        UniqueConstraint('hqpp_id', 'tbs_id', 'company_id')

        # FIXME: add PRN (Producer Registration Number)
        # FIXME: add status (pending / in_progress (by tbs) / completed (by tbs) )

    )

    tblWeMemberAccount = IQTable('tblWeMemberAccount', metadata,
        Column('account_id', Integer, ForeignKey('tblAccount.id', ondelete='CASCADE'), primary_key=True)
    )

    tblWeMemberAccountLpp = IQTable('tblWeMemberAccountLpp', metadata,
        Column('account_id', Integer, ForeignKey('tblWeMemberAccount.account_id', ondelete='CASCADE'), primary_key=True),
        Column('lpp_id', Integer, ForeignKey('tblWeLpp.id', ondelete='CASCADE'), primary_key=True),
    )


    tables['tblWeHqppLicense'] = tblWeHqppLicense
    tables['tblWeLpp'] = tblWeLpp

    tables['tblWeMemberAccount'] = tblWeMemberAccount
    tables['tblWeMemberAccountLpp'] = tblWeMemberAccountLpp

    __tables_defined = True

def do_mappings():
    """
    This function is called by Django middleware when needed.
    """
    global __mapping_defined
    if __mapping_defined:
        return

    from basesoft.web.contrib.account import models as account_models
    from basesoft.web.contrib.contact import models as contact_models
    from we.tbs import models as tbs_models

    mapper(member_models.WeHqppLicense, tables['tblWeHqppLicense'],
        polymorphic_identity='WeHqppLicense',
        inherits=license_models.License
    )

    mapper(member_models.WeLpp, tables['tblWeLpp'],
        properties = {
            'hqpp': relation(member_models.WeHqppLicense, backref='lpps'),
            'tbs': relation(tbs_models.WeTbsLicense),
            'company': relation(contact_models.Company, foreign_keys=tables['tblWeLpp'].c.company_id),
            'authorized_representative': relation(contact_models.Company, foreign_keys=tables['tblWeLpp'].c.ar_company_id),
            'invoice_company': relation(contact_models.Company, foreign_keys=tables['tblWeLpp'].c.invoice_company_id),
        }
    )

    mapper(member_models.WeMemberAccount, tables['tblWeMemberAccount'],
        properties = {
            'lpps': relation(member_models.WeLpp, secondary=tables['tblWeMemberAccountLpp'], order_by=member_models.WeLpp.name, passive_deletes=True),
        },
        polymorphic_identity='WeMemberAccount',
        inherits=account_models.Account
    )

    mapper(member_models.WeMemberAccountLpp, tables['tblWeMemberAccountLpp'],
        properties = {
            'account': relation(member_models.WeMemberAccount),
            'lpp': relation(member_models.WeLpp),
        }
    )

    mapper(member_models.WeCompany,
           inherits=contact_models.Company,
           polymorphic_identity='WeCompany'
    )

    __mapping_defined = True

def fill_data(conn):
    pass

def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'WEEE Europe member')

def app_permissions():
    from django.utils.translation import ugettext as _
    return [

        license_models.Permission(u'we_create_member_license', u"""
            This gives the right to create WE member licenses
            """, _("Create WE producer license")),

        license_models.Permission(u'we_view_member_license', u"""
            This gives the right to access WE member licenses
            """, _("View WE producer license")),

        license_models.Permission(u'we_manage_member_license', u"""
            This gives the right to access WE member licenses
            """, _("Manage WE producer license")),

        license_models.Permission(u'we_manage_lpp_companies', u"""
            This gives the right to manage WE local producer partners
            """, _("Manage WE LPP companies")),

        license_models.Permission(u'we_view_lpp_companies', u"""
            This gives the right to view WE local producer partners
            """, _("View WE LPP companies")),

        license_models.Permission(u'we_add_company', u"""
            This gives the right to create third party companies
            """, _("Create third party companies")),

    ]

def app_roles():
    from basesoft.web.contrib.notification.models import NotificationReminderMail
    return [
        {
            'name': 'WE Producer',
            'required_permission': 'we_manage_roles',
            'permissions': [
                'view_dashboard', 'save_dashboard', 'view_license',
                'create_file', 'view_file', 'view_notification', 'archive_notification',
                'we_view_member_license', 'we_manage_member_license',
                'we_view_lpp_companies',
                'we_view_index_widget',
                'we_create_products', 'we_view_products',
                'we_manage_declaration', 'we_view_declaration', 'we_view_declaration_order',
                'we_view_tbs_product_level',
                'view_datasource', 'view_datasource_widget', 'we_view_report',
            ],
            'license_configs': [],
            'account_configs': [
                ('basesoft.dashboard.compose.right_layout_unit.collapsed', 'True'),
                ('basesoft.dashboard.mode.value', 0),
                ('basesoft.account.use.iconbar', 'False'),
                ('basesoft.web.contrib.notification.reminder.mail', NotificationReminderMail.SEND_DAILY),
            ]
        }
    ]
