# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.

import os

import datetime

from datetime import date

from django.conf import settings
from django.template.defaultfilters import slugify
from django.utils.translation import ugettext as _

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy import and_, or_, func

from basesoft.web.core.exceptions import PublicUserException, PublicSecurityException
from basesoft.web.core.util.controllers import transaction_decorator, lazy_assert

from iqportal.weee.core.models import WEEEProductLevel

from we.core.controllers import AbstractWeController, WeNotificationController
from we.core.models import WeOperatorLicense
from we.member.models import WeLpp, WeHqppLicense
from we.product.models import WeCN8, WeUnuCode, WeHqppProduct, WeTbsProductLevel, \
    WeTbsRegulationPeriodProduct, WeHqppRegulationPeriodProduct, WeRegulationPeriodNotification, WeCN8PeriodProduct, \
    WeHqppProductExcludedAtTbs
from we.tbs.models import WeTbsRegulationPeriod, WeTbsLicense


class WeProductController(AbstractWeController):

    def get_weee_product_levels(self):
        options = []

        categories = self.get_privileged_object(WEEEProductLevel)

        for cat in categories:
            options.append({
                'label': '%s - %s' % (cat.label, cat.description),
                'value': cat.id
            })

        return options

    def get_cn8_codes(self, selected=None, filter_text='', limit=None, get_available=False):

        cn8 = self.sa_session.query(WeCN8)

        _filter = None

        if get_available:
            _filter = WeCN8.weee_product_level_id == None

        if filter_text != '':
            if _filter is None:
                _filter = or_(WeCN8.code.like('%%%s%%' % filter_text), WeCN8.description.like('%%%s%%' % filter_text))
            else:
                _filter = and_(_filter, or_(WeCN8.code.like('%%%s%%' % filter_text), WeCN8.description.like('%%%s%%' % filter_text)))

        if _filter is not None:
            cn8 = cn8.filter(_filter)

        if limit is not None:
            cn8 = cn8.limit(limit)

        rv = [{'label': str(cn.code) + ' - ' + cn.description, 'value': str(cn.code)} for cn in cn8.all()]

        if selected is not None and len(selected) > 0:
            [rv.append({'label': str(cn.code) + ' - ' + cn.description, 'value': str(cn.code)}) for cn in self.sa_session.query(WeCN8).filter(WeCN8.code.in_(selected)).all()]

        return rv

    def get_used_cn8_codes(self, product_level_id, cn8_codes=None):
        # get all cn8 codes not used in the provided weee subcategory
        _filter = WeCN8.weee_product_level_id != product_level_id

        if cn8_codes is not None:
            _filter = and_(_filter, WeCN8.code.in_(cn8_codes))

        return self.sa_session.query(WeCN8).filter(_filter).all()

    def get_unu_codes(self):
        rv = [{'label': unu.code, 'value': unu.id} for unu in self.sa_session.query(WeUnuCode).all()]
        return rv

    def get_cn8_by_product_level(self, product_level_id):
        return self.sa_session.query(WeCN8).filter(WeCN8.weee_product_level_id == product_level_id).all()

    @transaction_decorator
    def save_or_update_cn8_weee_product_level(self, _id, state):
        self.assert_permissions('we_manage_cn8')

        weee_product_level_id = int(state['weee_product_level_id'])
        state_codes = [int(code) for code in state['cn8_codes']]
        weee_product_level_codes = self.get_cn8_by_product_level(weee_product_level_id)

        if weee_product_level_id is not None:
            weee_product_level = self.get_privileged_object_by_id(WEEEProductLevel, weee_product_level_id).pop()
        else:
            raise PublicUserException(_('There is no such subcategory.'))

        # if db code in state_codes update else remove the link
        for code in weee_product_level_codes:
            if code in state_codes:
                state_codes.remove(code)
                code.weee_product_level = weee_product_level
                code.unu_code_id = state['unu_code']
            else:
                code.weee_product_level = None
            self.sa_session.add(code)

        # add the remaining codes
        for code in state_codes:
            cn8 = self.sa_session.query(WeCN8).filter(WeCN8.code == code).one()
            cn8.weee_product_level = weee_product_level
            cn8.unu_code_id = state['unu_code']
            self.sa_session.add(cn8)

        return weee_product_level

    def load_cn8_weee_product_level(self, product_level_id):
        self.assert_permissions('we_manage_cn8')

        cn8_codes = self.get_cn8_by_product_level(product_level_id)

        if len(cn8_codes) != 0:
            weee_product_level = cn8_codes[0].weee_product_level
            unu = cn8_codes[0].unu_code_id
        else:
            weee_product_level = self.get_privileged_object_by_id(WEEEProductLevel, weee_product_level_id).pop()  # WHATS THIS?
            unu = ''

        load_state = {
            'weee_product_level_id': weee_product_level.id,
            'cn8_codes': [code.code for code in cn8_codes],
            'unu_code': unu
        }

        return load_state

    def validate_cn8_codes(self, product_level_id, cn8_codes):
        used_codes = self.get_used_cn8_codes(product_level_id, cn8_codes)
        return [{'cn8_code': c.code, 'weee_product_level': c.weee_product_level.label} for c in used_codes]

    def import_cn8_codes(self, tmp_file, importer):
        """
        Use importer to insert/update CN8 codes linked to WEEE category and UNU codes from a file.
        """
        file_path = os.path.join(settings.TMP_FILE_REPOSITORY, tmp_file.file_hash)

        rv, failures = importer.import_file(file_path, self)
        self.logger.info('WEEE-CN8-UNU file imported')
        return failures

    @transaction_decorator
    def save_or_update_cn8_code(self, state):
        self.assert_permissions('we_manage_cn8')

        if state['cn8_code'] in ['', None]:
            raise PublicUserException("There is no CN8 code. Please provide one.")

        cn8 = self.sa_session.query(WeCN8).filter(WeCN8.code == state['cn8_code']).first()
        if cn8 is None:
            cn8 = WeCN8()
            cn8.code = state['cn8_code']
        cn8.description = state['cn8_description']
        self.sa_session.add(cn8)

        return cn8

    def save_or_update_tbs_product_level(self, _id, label, description, tbs_id, parent_level_id=None, invoice_basis=None,
                                         position=None, tbs_regulation_period_id=None, fee=None):
        """
        Used in product_level_details.
        """
        self.assert_permissions('we_manage_tbs_product_level')
        tbs = self.get_tbs_licenses(tbs_id).pop()
        return self.save_or_update_tbs_category(_id, label, description, tbs, parent_level_id, invoice_basis,
                                                position, tbs_regulation_period_id, fee)

    @transaction_decorator
    def save_or_update_tbs_category(self, _id, label, description, tbs, parent_level_id=None, invoice_basis=None,
                                         position=None, regulation_period_id=None, fee=None):

        # retrieve TBS for access check
        tbs = self.get_tbs_licenses(tbs.id).pop()

        p = position
        parent_level = None

        if _id is not None:
            tp = self.sa_session.query(WeTbsProductLevel).\
                filter(and_(WeTbsProductLevel.id == _id, WeTbsProductLevel.archived==0)).first()
            if tp is None:
                raise PublicSecurityException(_('No such take-back scheme local category.'))
            assert tp.tbs == tbs
            p = tp.position if position is None else position
        else:
            tp = WeTbsProductLevel()

        if invoice_basis is None:
            invoice_basis = WeTbsProductLevel.INVOICE_BASED_ON_WEIGHT

        # if we have a parent id supplied use it for a parent and check if the parent id is not a child
        # else use current parent
        if parent_level_id is not None:
            parent_level = self.sa_session.query(WeTbsProductLevel).\
                filter(and_(WeTbsProductLevel.id == parent_level_id,
                            WeTbsProductLevel.archived==0)).first()

            if tp.id is not None:
                self.product_level_has_child(tp, parent_level)

            # if parent level id is not the same change position
            if tp.parent_level is not None and tp.parent_level.id != parent_level_id:
                p = None

        elif tp.parent_level is not None:
            parent_level = self.sa_session.query(WeTbsProductLevel).\
                filter(and_(WeTbsProductLevel.id == tp.parent_level.id,
                            WeTbsProductLevel.archived==0)).first()

        if parent_level is not None and label == parent_level.label:
            raise PublicUserException(_('Subcategory %(sub)s cannot have the same label as its'
                                      ' main category %(main)s.' % {'sub': label, 'main': parent_level.label}))

        # get position when we are adding new product level
        if p is None and parent_level is not None and len(parent_level.sublevels) > 0:
            p = parent_level.sublevels[-1].position + 1
        elif p is not None:
            pass
        else:
            # get all levels with parent id None for the given tbs to get the position of next Main level
            levels = self.sa_session.query(WeTbsProductLevel).\
                filter(and_(WeTbsProductLevel.parent_level == parent_level,
                            WeTbsProductLevel.archived==0,
                            WeTbsProductLevel.tbs == tbs)).all()
            p = len(levels) + 1 if len(levels) > 0 else 1

        # if this is a new category no reason to reindex or if the position is the same.
        if tp.position is not None and tp.position != p:
            self._reindex_tbs_product_level(parent_level, p, tbs)
        tp.tbs = tbs
        tp.parent_level = parent_level
        tp.label = label
        tp.description = description
        tp.invoice_basis = invoice_basis
        tp.position = p
        self.sa_session.add(tp)

        # check whether regulation period exists
        if regulation_period_id is not None:
            regulation_period = self.sa_session.query(WeTbsRegulationPeriod).filter(
                and_(
                    WeTbsRegulationPeriod.id == regulation_period_id,
                    WeTbsRegulationPeriod.archived == 0,
                    WeTbsRegulationPeriod.tbs == tbs
                )
            ).first()

            if regulation_period is None:
                PublicUserException(_('Regulation period with ID %s does not exist or you have no access.'))

            if fee is not None and fee.is_signed() and not tbs.get_config('we.tbs.use.negative.fee'):
                raise PublicUserException(_("TBS %s does not allow negative fees." % tbs.name))

            # check whether the category is already linked to that regulation period
            regulation_period_product = self.sa_session.query(WeTbsRegulationPeriodProduct).filter(
                and_(
                    WeTbsRegulationPeriodProduct.tbs_product_level == tp,
                    WeTbsRegulationPeriodProduct.tbs_regulation_period == regulation_period
                )
            ).first()

            if regulation_period_product is None:
                # only new rpp should get a new position, this is only ever called through a manual addition of a new category
                # in case of an upload of a new pricesheet the position is done through save_or_update_regulation_period
                # for more information see ticket #649
                regulation_period_product = WeTbsRegulationPeriodProduct()
                regulation_period_product.tbs_product_level = tp

                # number of categories in regulation period
                p = len(regulation_period.period_products) + 1

                if parent_level is not None:
                    rpp = self.sa_session.query(WeTbsRegulationPeriodProduct).filter(
                        and_(
                            WeTbsRegulationPeriodProduct.tbs_product_level == parent_level,
                            WeTbsRegulationPeriodProduct.tbs_regulation_period == regulation_period
                        )
                    ).first()

                    # position of new product level
                    p = rpp.position + len(parent_level.sublevels)

                    # increment position of records with bigger position
                    q = self.sa_session.query(WeTbsRegulationPeriodProduct).filter(
                        and_(
                            WeTbsRegulationPeriodProduct.position >= p,
                            WeTbsRegulationPeriodProduct.tbs_regulation_period == regulation_period
                        )
                    ).order_by(WeTbsRegulationPeriodProduct.position.desc())

                    for pp in q:
                        pp.position += 1
                        self.sa_session.add(pp)
                        self.sa_session.flush()
                regulation_period_product.position = p

            regulation_period_product.tbs_regulation_period = regulation_period
            regulation_period_product.fee = fee
            self.sa_session.add(regulation_period_product)

        try:
            self.sa_session.flush()
        except IntegrityError, e:
            self.logger.info('Integrity error for duplicate on WeTbsProductLevel, %s' % e)
            raise PublicUserException(_('Duplicate entry for the provided category with label %(label)s' % {'label': label}))

        return tp

    @transaction_decorator
    def _reindex_tbs_product_level(self, parent_level, position, tbs):
        # get the child level with the given position
        pl = self.sa_session.query(WeTbsProductLevel)\
            .filter(and_(WeTbsProductLevel.parent_level == parent_level, WeTbsProductLevel.position == position)).first()

        if pl is None:
            return

        # if parent level is None get all main levels to get the position
        if parent_level is None:
            pls = self.sa_session.query(WeTbsProductLevel)\
                .filter(and_(WeTbsProductLevel.parent_level == None, WeTbsProductLevel.archived == 0,
                             WeTbsProductLevel.tbs == tbs)).all()
            p = pls[-1].position if len(pls) > 0 else 0
        else:
            p = pl.parent_level.sublevels[-1].position if len(pl.parent_level.sublevels) > 0 else 0
        # give it a new position at the end so we can save the other product level on that position
        pl.position = p + 1
        self.sa_session.add(pl)

    def product_level_has_child(self, product_level, new_parent):
        if product_level.id == new_parent.id:
            raise PublicUserException(_("Category %s cannot be a parent of itself." % new_parent.label))

        # current setup doesn't allow children to have children or parents to be children
        if new_parent.parent_level is not None:
            raise PublicUserException(_("Sub-categories cannot have sub-categories. Please choose a parent category which is not already a sub-category."))

        if len(product_level.sublevels) > 0:
            raise PublicUserException(_("Main categories cannot become sub-categories."))

    def get_tbs_product_levels(self, _id, tbs_id=None):
        self.assert_permissions('we_view_tbs_product_level')

        _filter = and_(WeTbsProductLevel.archived == 0, WeTbsProductLevel.deleted == 0)

        if type(_id) == list:
            return self.sa_session.query(WeTbsProductLevel).\
                filter(and_(_filter, WeTbsProductLevel.id.in_(_id))).all()

        if tbs_id is not None:
            _filter = and_(_filter, WeTbsProductLevel.tbs_id == tbs_id)
            # we cant join on regulation period product here as we do not have that scope.
            return self.sa_session.query(WeTbsProductLevel).order_by(WeTbsProductLevel.id).filter(_filter).all()

        return self.sa_session.query(WeTbsProductLevel).\
            filter(and_(_filter, WeTbsProductLevel.id == _id)).first()

    def get_tbs_product_levels_for_period(self, tbs_id, period=None, include_null_fees=True):
        if period is None:
            tbs = self.sa_session.query(WeTbsLicense).filter(WeTbsLicense.id==tbs_id).first()
            period = tbs.domain.today()
        rp = self.get_regulation_periods(tbs_id, start_date=period)
        if include_null_fees:
            return [p.tbs_product_level for p in rp.period_products]
        return [p.tbs_product_level for p in rp.period_products if p.fee is not None]

    def get_parent_category_for_tbs(self, tbs_id, regulation_period_id=None):
        self.assert_permissions('we_view_tbs_product_level')

        _filter = and_(WeTbsProductLevel.tbs_id == tbs_id, WeTbsProductLevel.parent_level == None,
                       WeTbsProductLevel.archived == 0, WeTbsProductLevel.deleted == 0)
        if regulation_period_id is not None:
            _filter = and_(_filter, WeTbsRegulationPeriodProduct.tbs_regulation_period_id == regulation_period_id)

        pls = self.sa_session.query(WeTbsProductLevel).filter(_filter).all()
        rv = []
        for level in pls:
            rv.append({
                'id': level.id,
                'label': level.label,
                'concat_label': '%s - %s' % (level.label, level.description),
                'description': level.description,
                'invoice_basis': level.invoice_basis,
            })

        return rv

    def get_product_details(self, ids, rp_id=None):
        _filter = and_(WeTbsProductLevel.archived == 0, WeTbsProductLevel.deleted == 0)
        _filter = and_(_filter, WeTbsProductLevel.id.in_(ids))

        levels = []
        if rp_id is not None:

            levels = self.sa_session.query(WeTbsProductLevel)\
                .outerjoin(WeTbsRegulationPeriodProduct, and_(WeTbsRegulationPeriodProduct.tbs_regulation_period_id == rp_id,
                                                              WeTbsRegulationPeriodProduct.tbs_product_level_id == WeTbsProductLevel.id))\
                .filter(_filter).order_by(WeTbsRegulationPeriodProduct.position).all()
        else:
            levels = self.sa_session.query(WeTbsProductLevel)\
                .filter(_filter).order_by(WeTbsProductLevel.id).all()

        return [{
            'id': p.id,
            'label': p.label,
            'description': p.description,
            'invoice_basis': p.invoice_basis
        } for p in levels]

    @transaction_decorator
    def save_or_update_hqpp_product(self, product_id, state, declaration_period=None):
        self.assert_permissions('we_create_products')

        hqpp = None
        hqpps = self.get_member_licenses(int(state['hqpp_id']))

        for hq in hqpps:
            if hq.id == int(state['hqpp_id']):
                hqpp = hq
            else:
                raise PublicSecurityException(_("No access to supplied HQPP."))

        if product_id is None:
            product = WeHqppProduct()
        else:
            product = self.sa_session.query(WeHqppProduct).filter(WeHqppProduct.id == product_id).one()
            assert product.hqpp == hqpp

        if state['unu_id'] in [None, '']:
            product.unu_id = None
        else:
            product.unu_id = state['unu_id']

        if state['hs'] in [None, '']:
            product.hs = None
        else:
            product.hs = state['hs']

        hqpp_name = hqpp.name
        # FIXME: enable the check if we get TBS users
        # tbs_product_level = self.sa_session.query(WeTbsProductLevel).filter(WeTbsProductLevel.id == state['tbs_product_level_id']).one()
        # if tbs_product_level.tbs.get_config('we.tbs.use.sales.price'):
        product.sales_price = state['sales_price']

        product.hqpp = hqpp
        product.name = state['product_name']
        product.sku = state['sku']
        product.cn8 = state['cn8_codes'][0]

        product.weight_per_unit = state['weight_per_unit']
        product.length_per_unit = state['length_per_unit']
        product.height_per_unit = state['height_per_unit']
        product.width_per_unit = state['width_per_unit']

        product.weight_cat1 = state['weight_cat1']
        product.weight_cat2 = state['weight_cat2']
        product.weight_cat3 = state['weight_cat3']
        product.weight_cat4 = state['weight_cat4']
        product.weight_documentation = state['weight_documentation']
        product.weight_package = state['weight_package']
        product.weight_batteries = state['weight_batteries']

        product.diagonal = state['diagonal']
        product.modulation = state['modulation']
        product.description = state['description']
        self.sa_session.add(product)

        try:
            self.sa_session.flush()
        except IntegrityError, e:
            self.logger.debug(e)
            raise PublicUserException(_('SKU %s has already been saved for the selected HQPP. '
                                        'Duplicate entries are not allowed.' % (state['sku'])))

        # TBS state can have length = 0, this way behavior between uploaded Product Sheet
        # and direct use of this method is the same.
        for row in state['tbs_rowcontrol']:
            is_b2b = WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE
            if row.get('is_b2b') not in (None, ''):
                is_b2b = row.get('is_b2b')
            delete = False

            q = self.sa_session.query(WeHqppProductExcludedAtTbs).filter(and_(
                WeHqppProductExcludedAtTbs.tbs_id == int(row['tbs_id']),
                WeHqppProductExcludedAtTbs.product == product,
                WeHqppProductExcludedAtTbs.is_b2b == is_b2b
            ))

            if row.get('is_excluded', False):
                entry = q.first()
                if entry is not None:
                    continue
                entry = WeHqppProductExcludedAtTbs()
                entry.product = product
                entry.tbs = self.sa_session.query(WeTbsLicense).filter(WeTbsLicense.id == int(row['tbs_id'])).one()
                entry.is_b2b = is_b2b
                self.sa_session.add(entry)
                self.logger.info('HQPP Product with SKU [%s] has been excluded at TBS [%s].' % (state['sku'], row['tbs_id']))

                # FIXME, delete previously set links??

                continue

            q.delete()

            # get related data
            if row['tbs_product_level_id'] not in [None, '']:
                pl = self.get_product_sheet_tbs_product_level(row['tbs_product_level_id'])
                tbs = pl.tbs
                assert int(row['tbs_id']) == tbs.id
                # if tbs product level is changed update the record
                if pl.tbs.get_config('we.tbs.require.end.user.distinction'):
                    is_b2b = row['is_b2b']

                # get the regulation period
                start_date = self.today()
                if declaration_period is not None:
                    start_date = declaration_period
                rp = self.get_regulation_periods(tbs.id, start_date=start_date)

                # get the tbs product level for the given regulation period
                trpp = self.sa_session.query(WeTbsRegulationPeriodProduct)\
                        .filter(and_(WeTbsRegulationPeriodProduct.tbs_product_level == pl,
                                     WeTbsRegulationPeriodProduct.tbs_regulation_period == rp)).first()

                # if there is no  tbs product level for that regulation period
                # user has provided wrong selection
                if trpp is None:
                    raise PublicUserException(_("There is no local category with ID %(id)s for regulation "
                                                "period %(period)s." % {'id': pl.id, 'period': '-' if rp is None else rp.title}))
            else:
                delete = True

            # add new or update
            if row['hqpp_period_product_id'] in [None, '']:

                if delete is False:
                    if trpp.fee is None:
                        raise PublicUserException(_('You cannot link a local category which does not have a fee set.'))
                    hrpp = WeHqppRegulationPeriodProduct()
                    hrpp.product = product
                    hrpp.tbs_regulation_period_product = trpp
                    hrpp.tbs_regulation_period = rp
                    hrpp.is_b2b = is_b2b
                    self.sa_session.add(hrpp)
                continue

            # if there is saved link and we have None for new tbs product level, remove the link
            if delete is True:
                # Delete - explicitly done manually due to SQLAlchemy
                # mapper issues with multiple relations referencing RegulationPeriod.
                self.sa_session.execute('DELETE FROM tblWeHqppRegulationPeriodProduct WHERE id = :id '
                                        'AND product_id = :product_id', {
                    'id': row['hqpp_period_product_id'],
                    'product_id': product.id
                })
                continue
            hrpp = self.sa_session.query(WeHqppRegulationPeriodProduct)\
                .filter(and_(WeHqppRegulationPeriodProduct.product == product,
                             WeHqppRegulationPeriodProduct.id == row['hqpp_period_product_id'])).one()

            # if there is no change go to next
            if hrpp.tbs_regulation_period_product.tbs_product_level == pl:
                continue

            if trpp.fee is None:
                raise PublicUserException(_('You cannot link a local category which does not have a fee set.'))

            hrpp.is_b2b = is_b2b
            hrpp.tbs_regulation_period_product = trpp
            self.sa_session.add(hrpp)

        self.logger.info('New HQPP Product with SKU [%s] has been saved for HQPP [%s]. ' % (state['sku'], hqpp_name))
        return product

    def get_product_sheet_tbs_product_level(self, id_obj):
        tbs_product_level_id = None
        if id_obj.has_key('suggested_categories'):
            tbs_product_level_id = id_obj['suggested_categories']
        else:
            tbs_product_level_id = id_obj['all_categories']
        return self.sa_session.query(WeTbsProductLevel).filter(WeTbsProductLevel.id == tbs_product_level_id).one()

    def get_suggested_tbs_product_levels(self, tbs_id, cn8, period, is_b2b=None):
        self.assert_permissions('we_view_tbs_product_level')

        if period is None:
            period = self.today()

        rp = self.get_regulation_periods(tbs_id, start_date=period)

        _filter = and_(WeTbsProductLevel.tbs_id == tbs_id,
                       WeTbsProductLevel.archived == 0, WeTbsProductLevel.deleted == 0,
                       WeCN8PeriodProduct.cn8_code == cn8,
                       WeTbsRegulationPeriodProduct.fee != None,
                       WeTbsRegulationPeriodProduct.tbs_regulation_period_id == rp.id)

        if is_b2b is None:
            is_b2b = WeCN8PeriodProduct.B2B_NOT_APPLICABLE
        _filter = and_(_filter, WeCN8PeriodProduct.is_b2b == is_b2b)
        return self.sa_session.query(WeTbsProductLevel).join(WeTbsRegulationPeriodProduct).join(WeCN8PeriodProduct)\
            .filter(_filter).order_by(WeTbsRegulationPeriodProduct.position).all()

    def get_hqpp_product(self, product_id):
        self.assert_permissions('we_view_products')

        group = self.get_default_group()
        product = self.sa_session.query(WeHqppProduct).filter(WeHqppProduct.id == product_id).one()
        # if current user active group.license.id != hqpp_id list will be empty
        hqpp = self.get_member_licenses(product.hqpp_id).pop()

        if isinstance(group.license, WeOperatorLicense):
            pass
        else:
            assert group.license == hqpp

        return product

    def load_hqpp_product(self, product_id):
        product = self.get_hqpp_product(product_id)

        hqpp = product.hqpp

        state = {
            'product_id': product.id,
            'product_name': product.name,
            'hqpp_id': hqpp.id,
            'hqpp_name': hqpp.company.name,
            'sku': product.sku,
            'cn8_codes': [product.cn8],
            'hs': product.hs,
            'unu_id': product.unu_id,

            'weight_per_unit': product.weight_per_unit,
            'length_per_unit': product.length_per_unit,
            'height_per_unit': product.height_per_unit,
            'width_per_unit': product.width_per_unit,
            'sales_price': product.sales_price,

            'description': product.description,

            'diagonal': product.diagonal,
            'modulation': product.modulation,

            'weight_cat1': product.weight_cat1,
            'weight_cat2': product.weight_cat2,
            'weight_cat3': product.weight_cat3,
            'weight_cat4': product.weight_cat4,
            'weight_documentation': product.weight_documentation,
            'weight_package': product.weight_package,
            'weight_batteries': product.weight_batteries,

        }

        return state

    def get_hqpp_product_tbs_links(self, member_id, product_id, declaration_period=None, cn8_code=None):
        options = []
        lpps = self.get_local_producer_partners(member_id=int(member_id), status=WeLpp.STATUS_ACTIVE)

        if not lpps:
            raise PublicUserException(_("There are no active LPPs at this HQPP. Please contact the operator."))

        tbs_ids = []

        for lpp in lpps:
            tbs = lpp.tbs
            if tbs is None or tbs.id in tbs_ids:
                continue

            tbs_ids.append(tbs.id)
            has_end_user_distinction = lpp.tbs.get_config('we.tbs.require.end.user.distinction')

            date_filter = self.today()

            if declaration_period is not None:
                date_filter = declaration_period

            # get the regulation period
            rp = self.get_regulation_periods(tbs.id, start_date=date_filter)

            if rp is None:
                options.append({
                    'tbs_name': tbs.name,
                    'error_msg': 'Take-back scheme does not have a regulation period '
                                 'for %s or earlier.' % date_filter
                })
                continue

            excluded = self.sa_session.query(WeHqppProductExcludedAtTbs).filter(and_(
                WeHqppProductExcludedAtTbs.product_id == product_id,
                WeHqppProductExcludedAtTbs.tbs == tbs
            )).order_by(WeHqppProductExcludedAtTbs.is_b2b).all()
            has_excluded = len(excluded) > 0

            # if TBS has b2b split get the links with is_b2b != 0 else get link with is_b2b == 0
            if not has_excluded and has_end_user_distinction:
                product_tbs_link = self.sa_session.query(WeHqppRegulationPeriodProduct)\
                    .join(WeTbsRegulationPeriodProduct).join(WeTbsProductLevel).\
                    filter(and_(WeHqppRegulationPeriodProduct.product_id == product_id,
                                WeHqppRegulationPeriodProduct.is_b2b != WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE,
                                WeTbsRegulationPeriodProduct.tbs_regulation_period_id == rp.id,
                                WeTbsProductLevel.tbs == tbs)).all()
            elif not has_excluded:
                product_tbs_link = self.sa_session.query(WeHqppRegulationPeriodProduct)\
                    .join(WeTbsRegulationPeriodProduct).join(WeTbsProductLevel).\
                    filter(and_(WeHqppRegulationPeriodProduct.product_id == product_id,
                                WeHqppRegulationPeriodProduct.is_b2b == WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE,
                                WeTbsRegulationPeriodProduct.tbs_regulation_period_id == rp.id,
                                WeTbsProductLevel.tbs == tbs)).all()
            else:
                product_tbs_link = []

            suggested = []

            if has_excluded or len(product_tbs_link) > 0:
                for p in product_tbs_link:
                    for category in p.tbs_products:
                        options.append({
                            'tbs_id': tbs.id,
                            'tbs_name': tbs.name,
                            'regulation_period_label': rp.title,
                            'hqpp_period_product_id': p.id,
                            'tbs_product_level_id': {'all_categories': category.id},
                            'tbs_product_level_label': '%s - %s' % (category.label, category.description),
                            'is_b2b': p.is_b2b,
                            'is_excluded': False
                        })
                        # we still need to render the empty row for B2B or B2C if only one is saved.
                        if len(product_tbs_link) == 1 and has_end_user_distinction:
                            is_b2b = WeHqppRegulationPeriodProduct.IS_B2B_FALSE
                            if WeHqppRegulationPeriodProduct.IS_B2B_FALSE == p.is_b2b:
                                is_b2b = WeHqppRegulationPeriodProduct.IS_B2B_TRUE

                            if cn8_code is not None:
                                suggested = self.get_suggested_tbs_product_levels(tbs.id, cn8_code, rp.start_date, is_b2b)

                            options.append({
                                'tbs_id': tbs.id,
                                'tbs_name': tbs.name,
                                'regulation_period_label': rp.title,
                                'is_b2b': is_b2b,
                                'is_excluded': False
                            })
                            if len(suggested) == 1:
                                options[-1]['suggested_category'] = '%s - %s' % (suggested[0].label, suggested[0].description)
                                options[-1]['tbs_product_level_id'] = {'suggested_categories': suggested[0].id}
                if has_excluded:
                    for ex in excluded:
                        options.append({
                            'tbs_id': tbs.id,
                            'tbs_name': tbs.name,
                            'regulation_period_label': rp.title,
                            'is_b2b': ex.is_b2b,
                            'is_excluded': True
                        })
                continue

            # get suggested category since we have no saved category
            if cn8_code is not None:
                if has_end_user_distinction is False:
                    suggested = self.get_suggested_tbs_product_levels(tbs.id, cn8_code, rp.start_date,
                                                                      WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE)
                    options.append({
                        'tbs_id': tbs.id,
                        'tbs_name': tbs.name,
                        'regulation_period_label': rp.title,
                        'is_b2b': WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE,
                        'is_excluded': False
                    })

                    if len(suggested) == 1:
                        options[-1]['suggested_category'] = '%s - %s' % (suggested[0].label, suggested[0].description)
                        options[-1]['tbs_product_level_id'] = {'suggested_categories': suggested[0].id}
                    continue
                for is_b2b in [WeHqppRegulationPeriodProduct.IS_B2B_TRUE, WeHqppRegulationPeriodProduct.IS_B2B_FALSE]:
                    suggested = self.get_suggested_tbs_product_levels(tbs.id, cn8_code, rp.start_date, is_b2b)
                    options.append({
                        'tbs_id': tbs.id,
                        'tbs_name': tbs.name,
                        'regulation_period_label': rp.title,
                        'is_b2b': is_b2b,
                        'is_excluded': False
                    })

                    if len(suggested) == 1:
                        options[-1]['suggested_category'] = '%s - %s' % (suggested[0].label, suggested[0].description)
                        options[-1]['tbs_product_level_id'] = {'suggested_categories': suggested[0].id}


        return options

    def get_skus(self, lpp_id=None, contains=None, exclude=None):
        self.assert_permissions('we_view_products')

        if not lpp_id:
            raise PublicUserException(_("Cannot find Local Producer Partner."))

        lpp = self.get_local_producer_partner(int(lpp_id))

        _where = and_(WeHqppProduct.hqpp_id == lpp.hqpp_id, WeTbsProductLevel.tbs == lpp.tbs)

        if contains is not None:
            _where = and_(_where, WeHqppProduct.sku.like('%%%s%%' % contains))

        if exclude is not None:
            _where = and_(_where, ~WeHqppProduct.sku.in_(exclude))

        skus = self.sa_session.query(WeHqppProduct).join(WeHqppRegulationPeriodProduct).join(WeTbsRegulationPeriodProduct).join(WeTbsProductLevel).filter(_where).all()
        return skus

    def get_product_info(self, lpp_id, sku, year, month):
        self.assert_permissions('we_view_products')

        lpp = self.get_local_producer_partner(lpp_id)
        b2b = WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE

        if lpp.tbs.get_config('we.tbs.require.end.user.distinction') is False:
            _filter = WeHqppRegulationPeriodProduct.is_b2b == b2b
        else:
            _filter = WeHqppRegulationPeriodProduct.is_b2b != b2b

        product = self.sa_session.query(WeHqppProduct).filter(and_(WeHqppProduct.hqpp_id == lpp.hqpp_id,
                                                                   WeHqppProduct.sku == sku)).first()
        period = date(int(year), int(month), 1)

        if product is None:
            return {
                'hqpp_id': lpp.hqpp.id,
                'js_module': 'we_hqpp_product_details',
                'declaration_period': period
            }

        rp = self.get_regulation_periods(lpp.tbs.id, start_date=period)
        if rp is None:
            raise PublicUserException(_('No regulation period available for the provided year and month.'))

        pp = self.sa_session.query(WeHqppRegulationPeriodProduct).filter(
            and_(
                WeHqppRegulationPeriodProduct.product==product,
                WeHqppRegulationPeriodProduct.tbs_regulation_period==rp,
                _filter
            )
        ).all()

        if len(pp) == 0:
            return {
                'hqpp_id': lpp.hqpp.id,
                'js_module': 'we_hqpp_product_details',
                'declaration_period': period,
                'product_id': product.id
            }

        weee_pl = None
        if product.cn8_code.weee_product_level:
            weee_pl = product.cn8_code.weee_product_level.label

        state = []

        for p in pp:
            fee = p.tbs_regulation_period_product.fee

            state.append({
                'product_name': product.name,
                'cn8': product.cn8,
                'weee_product_level': weee_pl,
                'weight_per_unit': product.weight_per_unit,
                'weight_calculation': product.get_total_weight_per_unit_for_regulation_period(rp),
                'tbs_product_level': p.tbs_regulation_period_product.tbs_product_level.label,
                'tbs_regulation_period': p.tbs_regulation_period.title,
                'is_b2b': p.is_b2b,
                'invoice_basis': p.tbs_regulation_period_product.tbs_product_level.invoice_basis,
                'fee': fee,
                'hqpp_period_product_id': p.id
            })

        if len(state) == 0:
            raise PublicUserException(_("There are no TBS local categories linked to product SKU %s" % sku))
        return state

    def import_cn8_to_tbs_product_level(self, tmp_file, importer, tbs_id, regulation_period_id):
        """
        Use importer to insert CN8 to TBS local category suggestions from a file.
        """
        self.assert_permissions('we_manage_cn8_tbs_product_level')

        rv, failures = importer.import_file(tmp_file, self, tbs_id=tbs_id, regulation_period_id=regulation_period_id)
        self.logger.info('CN8 to TBS product level file imported for TBS[%s]' % tbs_id)
        return failures

    def import_tbs_price_sheet(self, tmp_file, importer, tbs_id, row_count, regulation_period_id=None,
                               regulation_label=None, regulation_year=None, regulation_month=None, weight_settings=None):
        """
        Use importer to insert TBS categories from a file.
        """
        self.assert_permissions('we_manage_tbs_product_level')

        rv, failures = importer.import_file(tmp_file, self, tbs_id=tbs_id,
                                            regulation_period_id=regulation_period_id,
                                            regulation_label=regulation_label,
                                            regulation_year=regulation_year,
                                            regulation_month=regulation_month,
                                            row_count=row_count,
                                            weight_settings=weight_settings)
        self.logger.info('TBS product levels file imported for TBS[%s]' % tbs_id)
        return failures

    def import_hqpp_product_sheet(self, tmp_file, importer, hqpp_id, regulation_periods_state):
        """
        Use importer to insert/update HQPP product sheet from a file.
        """
        self.assert_permissions('we_create_products')
        file_path = os.path.join(settings.TMP_FILE_REPOSITORY, tmp_file.file_hash)

        rv, failures = importer.import_file(file_path, self, hqpp_id=hqpp_id,
                                            regulation_periods=regulation_periods_state)
        self.logger.info('Product sheet file imported for HQPP[%s]' % hqpp_id)
        return failures

    def export_hqpp_product_sheet(self, hqpp_id):
        """
        Writes data to excel file. User for Customdatatable data export.
        """
        self.assert_permissions('we_view_products')

        def __set_cell_value(sheet, value, row, column, data_type=None):
            c = sheet.cell(row=row, column=column)

            if data_type is not None:
                c.set_value_explicit(value)
                return
            c.value = value

        def __duplicate_file_row(sheet, headers, product, row):
            # first 16 columns are static after that are the TBS columns
            for k in xrange(0, 16):
                h = headers[k]
                if h == 'is_b2b':
                    continue
                __set_cell_value(sheet, getattr(product, h), row, k)

        headers = ['sku', 'name', 'description', 'weight_per_unit', 'length_per_unit', 'height_per_unit', 'width_per_unit',
                   'diagonal', 'weight_cat1', 'weight_cat2', 'weight_cat3', 'weight_cat4', 'weight_documentation', 'weight_batteries',
                   'weight_package', 'cn8', 'is_b2b']
        user_headers = [_('SKU'), _('Product name'), _('Description'), _('Weight per unit (kg)'), _('Length per unit (cm)'),
                        _('Height per unit (cm)'), _('Width per unit (cm)'), _('Diagonal (inches)'), _('Weight category 1(kg)'),
                        _('Weight category 2 (kg)'), _('Weight category 3 (kg)'), _('Weight category 4 (kg)'), _('Weight documentation (kg)'),
                        _('Weight batteries (kg)'), _('Weight package (kg)'), _('CN8 code'), _('B2B/B2C')]
        tbs_configs = {}

        lazy_assert(hqpp_id, int)
        # get all TBS at which HQPP is registered and active.
        member = self.get_member_licenses(hqpp_id).pop()
        lpps = self.get_local_producer_partners(member.id)
        tbses = [lpp.tbs for lpp in lpps]

        for tbs in tbses:
            # no duplicate TBSes
            if tbs_configs.has_key(tbs.license_id):
                continue
            headers.append('tbs-%d' % tbs.license_id)
            user_headers.append(tbs.company.name)
            tbs_configs[tbs.license_id] = tbs
        # get all products of hqpp
        products = self.sa_session.query(WeHqppProduct).filter(WeHqppProduct.hqpp_id == hqpp_id)

        # create xlsx file
        import openpyxl.workbook as workbook
        book = workbook.Workbook()
        sheet = book.get_active_sheet()
        sheet.append(headers)
        sheet.append(user_headers)

        current_row = 2
        i = 0
        # get current regulation period
        regulation_periods = []
        for tbs in tbses:
            rp = self.get_regulation_periods(tbs.id, start_date=self.today())
            if rp is not None:
                regulation_periods.append(rp.id)
        # for every product add row and check for is_b2b
        for product in products:
            no_distinction = []
            b2b = []
            b2c = []
            b2b_tbses = []
            b2c_tbses = []
            no_b2b_tbses = []
            b2b_row = current_row
            b2c_row = b2b_row + 1
            product_has_distinction = False

            _filter = and_(WeHqppRegulationPeriodProduct.product_id == product.id,
                           WeHqppRegulationPeriodProduct.tbs_regulation_period_id.in_(regulation_periods))
            links = self.sa_session.query(WeHqppRegulationPeriodProduct).filter(_filter).all()

            for link in links:
                linked_tbs = link.tbs_regulation_period_product.tbs_product_level.tbs

                if linked_tbs in tbses:
                    if linked_tbs.get_config('we.tbs.require.end.user.distinction') is False:
                        if link.is_b2b == WeHqppRegulationPeriodProduct.B2B_NOT_APPLICABLE:
                            no_b2b_tbses.append(linked_tbs.id)
                            no_distinction.append(link)
                    else:
                        if link.is_b2b == WeHqppRegulationPeriodProduct.IS_B2B_FALSE:
                            b2c_tbses.append(linked_tbs.id)
                            b2c.append(link)
                        elif link.is_b2b == WeHqppRegulationPeriodProduct.IS_B2B_TRUE:
                            b2b_tbses.append(linked_tbs.id)
                            b2b.append(link)

            for j in xrange(0, len(headers)):
                header = headers[j]

                if header == 'is_b2b':
                    if len(b2b) > 0:
                        __set_cell_value(sheet, 'B2B', i + b2b_row, j)
                        __duplicate_file_row(sheet, headers, product, i + b2b_row)
                    if len(b2c) > 0:
                        product_has_distinction = True
                        __set_cell_value(sheet, 'B2C', i + b2c_row, j)
                        __duplicate_file_row(sheet, headers, product, i + b2c_row)
                    continue

                if header.startswith('tbs-'):

                    tbs_id = int(header[4:])

                    # if TBS has end user distinction we add two rows b2b and b2c
                    if tbs_configs[tbs_id].get_config('we.tbs.require.end.user.distinction'):
                        if tbs_id in b2b_tbses:
                            ind = b2b_tbses.index(tbs_id)
                            tbs_product_level = b2b[ind].tbs_regulation_period_product.tbs_product_level
                            __set_cell_value(sheet, tbs_product_level.label, i + b2b_row, j, data_type='s')

                        elif tbs_id in b2c_tbses:
                            ind = b2c_tbses.index(tbs_id)
                            tbs_product_level = b2c[ind].tbs_regulation_period_product.tbs_product_level
                            __set_cell_value(sheet, tbs_product_level.label, i + b2c_row, j, data_type='s')
                    else:
                        if tbs_id in no_b2b_tbses:
                            ind = no_b2b_tbses.index(tbs_id)
                            tbs_product_level = no_distinction[ind].tbs_regulation_period_product.tbs_product_level
                            __set_cell_value(sheet, tbs_product_level.label, i + current_row, j, data_type='s')

                    if product_has_distinction and j >= len(headers) - 1:
                        current_row = b2c_row
                    continue

                __set_cell_value(sheet, getattr(product, headers[j]), i + current_row, j)
            i += 1

        f_name = os.path.join(settings.TMP_FILE_REPOSITORY, '%s_product_sheet.xlsx' % slugify(member.name))
        book.save(f_name)

        return f_name

    @transaction_decorator
    def delete_cn8_code_matches(self, codes):
        self.assert_permissions('we_manage_cn8')

        for c in codes:
            code = self.sa_session.query(WeCN8).filter(WeCN8.code == int(c)).one()
            code.unu_code = None
            code.weee_product_level = None
            self.sa_session.add(code)

    @transaction_decorator
    def mark_tbs_product_levels_as_archived(self, obj, product_levels):
        self.assert_permissions('we_manage_tbs_product_level')

        no_access = []
        has_rp = []
        tbs = self.get_tbs_licenses()
        if len(tbs) == 1:
            tbs = tbs.pop()

        for _id in product_levels:
            try:
                # if we have multiple license this is operator account
                if type(tbs) == list:
                    l = self.sa_session.query(WeTbsProductLevel).filter(WeTbsProductLevel.id == int(_id)).one()
                else:
                    l = self.sa_session.query(WeTbsProductLevel).filter(
                        and_(
                            WeTbsProductLevel.tbs == tbs,
                            WeTbsProductLevel.id == int(_id)
                        )
                    ).one()
                rpp = self.sa_session.query(WeTbsRegulationPeriodProduct).filter(
                    WeTbsRegulationPeriodProduct.tbs_product_level == l
                ).first()

                if rpp is not None:
                    has_rp.append(l.label)
                    continue
                l.archived = True
                self.sa_session.add(l)
            except NoResultFound:
                no_access.append(_id)

        if len(no_access) > 0:
            s = ', '.join(no_access)
            raise PublicUserException(_("You have no access to categories %s." % s))
        if len(has_rp) > 0:
            raise PublicUserException(_("Categories which are part of a regulation period cannot be archived."
                                        " Only whole regulation periods can be archived."))
        return []

    def get_regulation_periods(self, tbs_id, label=None, start_date=None):
        if isinstance(tbs_id, WeTbsLicense):
            tbs = tbs_id
        else:
            tbs = self.get_tbs_licenses(tbs_id).pop()

        _filter = and_(WeTbsRegulationPeriod.tbs == tbs, WeTbsRegulationPeriod.archived == 0)

        if label is not None:
            _filter = and_(_filter, WeTbsRegulationPeriod.label == label)

        if start_date is not None:
            _filter = and_(_filter, WeTbsRegulationPeriod.start_date <= start_date)
            return self.sa_session.query(WeTbsRegulationPeriod).filter(_filter)\
                .order_by(WeTbsRegulationPeriod.start_date.desc()).first()

        return self.sa_session.query(WeTbsRegulationPeriod).filter(_filter).all()

    def get_regulation_period_by_id(self, tbs_id, regulation_period_id):
        rp = self.sa_session.query(WeTbsRegulationPeriod).filter(WeTbsRegulationPeriod.id == regulation_period_id).one()
        if tbs_id is not None:
            assert rp.tbs.id == tbs_id
        # used for validation of access
        tbs = self.get_tbs_licenses(rp.tbs.id).pop()
        return rp

    def get_available_months_in_year(self, tbs_id, year):
        # access validation
        tbs = self.get_tbs_licenses(tbs_id).pop()
        rps = self.sa_session.query(WeTbsRegulationPeriod).filter(
            and_(
                WeTbsRegulationPeriod.tbs_id == tbs.id,
                WeTbsRegulationPeriod.archived == 0,
                func.year(WeTbsRegulationPeriod.start_date) == year
            )
        )

        months = range(1, 13)

        for rp in rps:
            months.remove(rp.start_date.month)

        _map = self.get_months_map()
        rv = []

        for i in months:

            rv.append({
                'value': i,
                'label': _map[i - 1]
            })

        return rv

    @transaction_decorator
    def save_or_update_regulation_period(self, _id, tbs_id, label, start_year, start_month, weight_settings,
                                         local_categories, return_errors=False):
        self.assert_permissions('we_manage_tbs_product_level')
        lazy_assert(local_categories, list)
        assert start_year > 1999  # we shouldn't go back too far ;)

        start_date = datetime.date(start_year, start_month, 1)

        tbs = self.get_tbs_licenses(tbs_id).pop()
        if _id is None:
            self.logger.info('Adding a new Regulation Period for TBS %s for period %s' % (tbs, start_date))

            rp = WeTbsRegulationPeriod()
            rp.tbs = tbs

            orp = self.sa_session.query(WeTbsRegulationPeriod).filter(
                and_(
                    WeTbsRegulationPeriod.tbs == tbs,
                    WeTbsRegulationPeriod.start_date == start_date,
                    WeTbsRegulationPeriod.archived == 0
                )
            ).first()
            if orp is not None:
                raise PublicUserException(_('There is already a regulation period for %s at this TBS.' % start_date))

        else:
            rp = self.sa_session.query(WeTbsRegulationPeriod).filter(
                and_(
                    WeTbsRegulationPeriod.id == _id,
                    WeTbsRegulationPeriod.archived == 0
                )
            ).one()
            assert rp.tbs is tbs
            self.logger.info('Updating Regulation Period (%s) for TBS %s' % (rp.label, rp.tbs))
            if rp.start_date != start_date:
                if self.today() <= start_date:
                    raise PublicUserException(_('The start of this Regulation Period cannot be in the past.'))
                self.logger.info(' - Updating start date %s' % start_date)

        # TODO, more verifications on start dates:
        #   declarations for requested period are already done and linked to another regulation period.
        rp.start_date = start_date
        rp.label = label
        rp.weight_settings = weight_settings

        lc = {}
        category_positions = {}
        category_dict = {}
        pos = 1

        error_list = []

        for p in local_categories:
            if p['fee'] not in [None, ''] and p['fee'].is_signed() and not tbs.get_config('we.tbs.use.negative.fee'):
                raise PublicUserException(_("TBS %s does not allow negative fees." % tbs.name))
            lc[int(p['id'])] = None if p['fee'] in ['', None] else p['fee']
            category_positions[int(p['id'])] = pos
            pos += 1

            if return_errors is True:
                category_dict[int(p['id'])] = {
                    'line_number': p['line_number'],
                    'row': p['row']
                }

        products = self.get_tbs_product_levels(lc.keys())
        for p in products:
            if p.tbs != tbs:
                raise PublicUserException(_('Incorrect local category for TBS: %s' % p.label))

        # set positions for current period_products to prevent integrity errors
        increment = len(local_categories) + 1
        for pp in rp.period_products:
            pp.position += increment
            self.sa_session.add(pp)
        self.sa_session.flush()

        for pp in rp.period_products:
            # the TBS product is no longer selected for this regulation period.
            if pp.tbs_product_level not in products:
                # check if the product level is linked to hqpp product
                link = self.sa_session.query(WeHqppRegulationPeriodProduct).filter(
                    WeHqppRegulationPeriodProduct.tbs_regulation_period_product_id == pp.id
                ).first()
                if link is not None:
                    message = _("Category %s cannot be removed from the regulation period because "
                                "it is linked to a producer product. If the description has changed, "
                                "do it manually." % pp.tbs_product_level.label)
                    if return_errors is False:
                        raise PublicUserException(message)
                    error_list.insert(0, [''] + [message])
                    continue
                # this will trigger an exception in case it is already used by a member.
                self.logger.info(' - Removing local category %s from Regulation Period' % pp.tbs_product_level)
                self.sa_session.delete(pp)
                # flush the current state to prevent duplicate entries on position.
                self.sa_session.flush()
                continue

            if pp.fee != lc[pp.tbs_product_level.id]:
                self.logger.info(' - Setting new fee for local category %s' % pp.tbs_product_level)
                pp.fee = lc[pp.tbs_product_level.id]
            pp.position = category_positions[pp.tbs_product_level_id]

        if _id is not None:
            self.sa_session.expire(rp)

        new_products = []

        for p in products:
            cat_key = (p.label, p.parent_level_id)
            if cat_key in new_products:
                message = _("You cannot have two categories with the same label '%(label)s' "
                            "and the same parent category ID '%(id)s' "
                            "in the same regulation period." % {'label': p.label, 'id': p.parent_level_id})

                if return_errors is False:
                    raise PublicUserException(message)
                error_list.append([category_dict[int(p.id)]['line_number']] +
                                  category_dict[int(p.id)]['row'] + [message])
                continue
            new_products.append(cat_key)
            if p in [prod.tbs_product_level for prod in rp.period_products]:
                continue

            pp = WeTbsRegulationPeriodProduct()
            pp.tbs_product_level = p
            pp.position = category_positions[p.id]
            pp.fee = lc[p.id]
            self.logger.info(' - Adding local category %s to Regulation Period' % pp.tbs_product_level)
            rp.period_products.append(pp)

        self.sa_session.add(rp)

        if return_errors is True and len(error_list) > 0:
            raise PublicUserException('Errors found during save of regulation period.', error_list)

        return rp

    def load_regulation_period_state(self, _id):
        rp = self.sa_session.query(WeTbsRegulationPeriod).filter(WeTbsRegulationPeriod.id == _id).first()

        # validate access to TBS
        self.get_tbs_licenses(rp.tbs.id).pop()

        pl = []
        lc = {}
        disabled = False
        # check if that regulation period is linked to hqpp products - then we cannot archive
        hrpp = self.sa_session.query(WeHqppRegulationPeriodProduct).filter(
            WeHqppRegulationPeriodProduct.tbs_regulation_period == rp).first()
        if hrpp is not None:
            disabled = True
        for p in rp.period_products:
            pl.append(p.tbs_product_level.id)
            lc['fee_%d' % p.tbs_product_level.id] = p.fee

        return {
            'period': {
                'id': _id,
                'tbs_id': rp.tbs.id,
                'label': rp.label,
                'start_year': rp.start_date.year,
                'start_month': rp.start_date.month,
                'tbs_product_levels': pl,
                'weight_settings': rp.weight_settings,
                'is_disabled': disabled
            },
            'local_categories': lc
        }

    def get_producer_regulation_period_products(self, hqpp, period):
        # access validation
        if self.get_account() is not None:
            self.get_member_licenses(hqpp.id)
        return self.sa_session.query(WeHqppRegulationPeriodProduct).join(WeHqppProduct).join(WeTbsRegulationPeriodProduct).filter(
            and_(
                WeTbsRegulationPeriodProduct.tbs_regulation_period == period,
                WeHqppProduct.hqpp == hqpp
            )
        )

    @transaction_decorator
    def save_or_update_hqpp_regulation_period_product(self, product_id, regulation_period_id, tbs_product_level_id, is_b2b):
        product = self.get_hqpp_product(product_id)
        regulation_period = self.get_regulation_period_by_id(None, regulation_period_id)
        product_level = self.get_tbs_product_levels(tbs_product_level_id)
        tbs_regulation_period_product = [p for p in regulation_period.period_products if p.tbs_product_level == product_level].pop()

        hrpp = self.sa_session.query(WeHqppRegulationPeriodProduct).filter(
            and_(
                WeHqppRegulationPeriodProduct.product == product,
                WeHqppRegulationPeriodProduct.tbs_regulation_period == regulation_period,
                WeHqppRegulationPeriodProduct.is_b2b == is_b2b
            )
        ).first()
        if hrpp is not None:
            if tbs_regulation_period_product == hrpp.tbs_regulation_period_product.tbs_product_level:
                return hrpp
            self.logger.info('Changing product level for regulation period from %s to %s' % (
                hrpp.tbs_regulation_period_product,
                tbs_regulation_period_product
            ))
        else:
            hrpp = WeHqppRegulationPeriodProduct()
            hrpp.product = product
            hrpp.tbs_regulation_period = regulation_period
            hrpp.is_b2b = is_b2b
        hrpp.tbs_regulation_period_product = tbs_regulation_period_product
        self.sa_session.add(hrpp)
        return hrpp

    @transaction_decorator
    def compare_regulation_period_with_previous(self, _id, dry_run=True):
        # on upload of a new price sheet, we only use dry run to show the operator/tbs what is going to happen on period change

        if self.get_account() is not None:
            self.assert_permissions('we_manage_tbs_product_level')

        if isinstance(_id, WeTbsRegulationPeriod):
            cp = _id
        else:
            cp = self.sa_session.query(WeTbsRegulationPeriod).filter(
                and_(
                    WeTbsRegulationPeriod.id == _id,
                    WeTbsRegulationPeriod.archived == 0
                )
            ).first()
        self.logger.info('Comparing Regulation Period %s with previous period.' % cp.title)

        # get the previous period
        pp = self.sa_session.query(WeTbsRegulationPeriod).filter(
            and_(
                WeTbsRegulationPeriod.start_date < cp.start_date,
                WeTbsRegulationPeriod.tbs == cp.tbs,
                WeTbsRegulationPeriod.archived == 0
            )
        ).order_by(WeTbsRegulationPeriod.start_date.desc()).first()

        if pp is None:
            self.logger.info(' - No previous period found, nothing to compare.')
            return None

        self.logger.info(' - Previous period %s' % pp.title)

        _map = {}

        # make a list of tbs products which are:
        #  * in the previous period, but no longer in the new one
        #  * check if they have the same label as one in the new period (description changed).
        for period_product in pp.period_products:
            not_found = True
            for new_pp in cp.period_products:

                if period_product.tbs_product_level == new_pp.tbs_product_level:
                    _map[period_product] = ('nochange', new_pp, )
                    not_found = False
                    break

                if period_product.tbs_product_level.label.lower() == new_pp.tbs_product_level.label.lower():
                    _map[period_product] = ('modified', new_pp, )
                    not_found = False
                    break

            if not_found:
                _map[period_product] = ('removed', None, )

        # get all producers who have those products linked
        producers = self.sa_session.query(WeHqppLicense).join(WeHqppProduct).join(WeHqppRegulationPeriodProduct).filter(
            WeHqppRegulationPeriodProduct.tbs_regulation_period_product_id.in_([p.id for p in pp.period_products]))

        affected = []
        automatic = []

        nc = WeNotificationController(self.sa_session)

        # loop producers
        for producer in producers:
            # cant use get_local_producer_partners here as that does security checks and this is accessed from Task.
            lpp = self.sa_session.query(WeLpp).filter(
                and_(
                    WeLpp.hqpp == producer,
                    WeLpp.tbs == cp.tbs,
                    WeLpp.status == WeLpp.STATUS_ACTIVE
                )
            ).first()

            if lpp is None:
                self.logger.info(' - Skipping producer %s due to inactive LPP' % producer)
                continue

            notification_params = []

            # get all producer products for previous period.
            for producer_product in self.get_producer_regulation_period_products(producer, pp):
                action, target = _map[producer_product.tbs_regulation_period_product]

                # if not dry_run: set local categories with links to the SKU where possible.
                if action in ('removed', 'modified',):
                    if producer not in affected:
                        affected.append(producer)
                    self.logger.info(' - Product (%s) can\'t be updated as the period product has been %s.' %
                                     (producer_product.tbs_regulation_period_product, action))
                    notification_params.append((producer_product.product.id, producer_product.product.sku))

                else:
                    self.logger.info(' - Product (%s) can be automatically updated to new regulation period product.' %
                                     producer_product.tbs_regulation_period_product)
                    if dry_run:
                        continue
                    prpp = WeHqppRegulationPeriodProduct()
                    prpp.product = producer_product.product
                    prpp.is_b2b = producer_product.is_b2b
                    prpp.tbs_regulation_period_product = target
                    prpp.tbs_regulation_period = target.tbs_regulation_period
                    self.sa_session.add(prpp)

            if producer not in affected:
                automatic.append(producer)

            # if not dry run: compile a notification which mentions which products were not automatically linked + reason + corben
            if not dry_run and len(notification_params) > 0:

                callback = {
                    'js_module': 'we_hqpp_new_regulation_period_wizard',
                    'parameters': [{
                        'products': notification_params,
                        'tbs': cp.tbs.name,
                        'regulationPeriodTitle': cp.title,
                        'regulationPeriodID': cp.id
                    }]
                }

                notification = nc.set_notification('New regulation period - action required', '',
                                    instance_class=WeRegulationPeriodNotification, callback=callback)
                nc.create_system_notification(notification, producer.default_group, for_license=True)

        if dry_run:
            rv = {
                'previous_period': pp.title,
                'removed': [],
                'modified': [],
                'nochange': [],
                'affected': [p.name for p in affected],
                'automatic': [p.name for p in automatic],
            }
            for previous, result in _map.iteritems():
                k, v = result
                rv[k].append((previous.get_state(), v if v is None else v.get_state()))
            return rv

        cp.is_processed = True
        self.sa_session.add(cp)

        return True

    @transaction_decorator
    def archive_regulation_period(self, rp_id):
        self.assert_permissions('we_manage_tbs_product_level')

        rp = self.sa_session.query(WeTbsRegulationPeriod).filter(WeTbsRegulationPeriod.id == rp_id).first()

        # validate access to TBS
        self.get_tbs_licenses(rp.tbs.id).pop()

        # check if that regulation period is linked to hqpp products - then we cannot archive
        hrpp = self.sa_session.query(WeHqppRegulationPeriodProduct).filter(
            WeHqppRegulationPeriodProduct.tbs_regulation_period == rp).first()

        if hrpp is not None:
            return False

        # archive
        rp.archived = True
        self.sa_session.add(rp)

        return True

    def _order_tbs_product_levels(self, product_levels):

        if len(product_levels) == 0:
            return []

        _map = {}
        for p in product_levels:
            _map[p.id] = p

        sql = """
                SELECT p.id,
                CASE
                    WHEN c.position IS NULL THEN p.position * 1000
                    ELSE p.position + c.position * 1000
                END AS ordering
                FROM tblWeTbsProductLevel AS p
                LEFT JOIN tblWeTbsProductLevel AS c
                    ON p.parent_level_id = c.id
                WHERE p.id IN (%s)
                ORDER BY ordering
            """ % ', '.join([str(p) for p in _map.keys()])

        rv = []

        result = self.sa_session.execute(sql)
        for _id, order in result.fetchall():
            rv.append(_map[_id])

        assert len(rv) == len(product_levels)

        return rv
