# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.

import datetime

from sqlalchemy import Integer, Column, UnicodeText, DateTime, ForeignKey, SmallInteger, Date

from basesoft.web.contrib.license import models as license_models
from basesoft.sqlalchemy.table import IQTable
from sqlalchemy.orm import mapper, relation, backref
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy.types import String, Numeric, Unicode
from we.product.models import WeTbsProductLevel

__mapping_defined = False
__tables_defined = False
tables = {}

def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False

def get_tables():
    return tables

def do_tables(metadata):
    """
    This function is called by Django middleware when needed.
    """
    global __tables_defined
    if __tables_defined:
        return

    # holds all possible CN8 codes
    tblWeCN8 = IQTable('tblWeCN8', metadata,
        Column('code', Integer, nullable=False, primary_key=True, autoincrement=False),
        Column('unu_code_id', Integer, ForeignKey('tblWeUnuCode.id', ondelete='RESTRICT')),
        Column('weee_product_level_id', Integer, ForeignKey('tblWEEEProductLevel.id', ondelete='RESTRICT')),
        Column('description', UnicodeText, nullable=False),

        Column('created', DateTime(), nullable=True, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),
    )

    tblWeUnuCode = IQTable('tblWeUnuCode', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('code', String(10), nullable=False),
        Column('description', UnicodeText, nullable=False),
    )

    tblWeTbsProductLevel = IQTable('tblWeTbsProductLevel', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('tbs_id', Integer, ForeignKey('tblWeTbsLicense.license_id'), nullable=False),
        Column('parent_level_id', ForeignKey('tblWeTbsProductLevel.id')),
        Column('position', Integer, nullable=False),
        Column('invoice_basis', SmallInteger, nullable=False, default=WeTbsProductLevel.INVOICE_BASED_ON_WEIGHT),

        Column('label', Unicode(255), nullable=False),
        Column('description', UnicodeText, nullable=False),

        Column('deleted', SmallInteger, nullable=False, default=0),
        Column('archived', SmallInteger, nullable=False, default=0),

        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),

        UniqueConstraint('parent_level_id', 'position')
    )

    tblWeTbsRegulationPeriodProduct = IQTable('tblWeTbsRegulationPeriodProduct', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('tbs_product_level_id', Integer, ForeignKey('tblWeTbsProductLevel.id', ondelete='CASCADE'), nullable=False),
        Column('tbs_regulation_period_id', Integer, ForeignKey('tblWeTbsRegulationPeriod.id', ondelete='CASCADE'), nullable=False),
        Column('position', Integer, nullable=False),
        Column('fee', Numeric(10, 5), nullable=True, default=None),

        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),

        UniqueConstraint('tbs_product_level_id', 'tbs_regulation_period_id'),
        UniqueConstraint('tbs_regulation_period_id', 'position')
    )

    # holds all CN8 to local category links as imported (used for suggestions on product sheet)
    tblWeCN8PeriodProduct = IQTable('tblWeCN8PeriodProduct', metadata,
        Column('id', Integer, nullable=False, primary_key=True, autoincrement=True),
        Column('cn8_code', Integer, ForeignKey('tblWeCN8.code', ondelete='RESTRICT'), nullable=False),
        Column('period_product_id', Integer, ForeignKey('tblWeTbsRegulationPeriodProduct.id', ondelete='CASCADE'), nullable=False),
        Column('is_b2b', SmallInteger, nullable=False, server_default='0'), # 0 = Not supplied, 1 = b2c, 2 = b2b

        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),

        UniqueConstraint('cn8_code', 'period_product_id', 'is_b2b')
    )

    # tblWeHqppProduct contains the generic product information
    # applicable to declarations from lpps at this hqpp to all tbs's
    tblWeHqppProduct = IQTable('tblWeHqppProduct', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('hqpp_id', Integer, ForeignKey('tblWeHqppLicense.license_id', ondelete='CASCADE'), nullable=False),
        Column('sku', String(20), nullable=False),
        Column('name', UnicodeText, nullable=False),
        Column('cn8', Integer, ForeignKey('tblWeCN8.code', ondelete='RESTRICT'), nullable=False),
        Column('unu_id', Integer, ForeignKey('tblWeUnuCode.id', ondelete='CASCADE')),  # optional
        Column('hs', Integer),  # optional

        Column('weight_per_unit', Numeric(15, 3)),  # in kg
        Column('length_per_unit', Numeric(15, 3)),  # in cm
        Column('height_per_unit', Numeric(15, 3)),  # in cm
        Column('width_per_unit', Numeric(15, 3)),  # in cm
        Column('sales_price', Numeric(10, 5)), # used at Swiss TBSes

        Column('weight_cat1', Numeric(7, 3)),  # in kg
        Column('weight_cat2', Numeric(7, 3)),  # in kg
        Column('weight_cat3', Numeric(7, 3)),  # in kg
        Column('weight_cat4', Numeric(7, 3)),  # in kg
        Column('weight_documentation', Numeric(7, 3)),  # in kg
        Column('weight_batteries', Numeric(7, 3)),  # in kg
        Column('weight_package', Numeric(7, 3)),  # in kg

        Column('diagonal', Numeric(7, 2)),  # inches
        Column('modulation', Unicode(5)),  # eco-friendly criteria

        Column('description', UnicodeText),

        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),

        UniqueConstraint('hqpp_id', 'sku')
    )

    # table linking a product to a local category
    tblWeHqppRegulationPeriodProduct = IQTable('tblWeHqppRegulationPeriodProduct', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('product_id', Integer, ForeignKey('tblWeHqppProduct.id', ondelete='RESTRICT'), nullable=False),
        Column('tbs_regulation_period_product_id', Integer, ForeignKey('tblWeTbsRegulationPeriodProduct.id', ondelete='RESTRICT'), nullable=False),
        Column('tbs_regulation_period_id', Integer, ForeignKey('tblWeTbsRegulationPeriod.id', ondelete='RESTRICT'), nullable=False),
        Column('is_b2b', SmallInteger, nullable=False, server_default='0'),  # 0 = Not supplied, 1 = b2c, 2 = b2b

        UniqueConstraint('product_id', 'tbs_regulation_period_id', 'is_b2b')
    )

    tblWeHqppProductExcludedAtTbs = IQTable('tblWeHqppProductExcludedAtTbs', metadata,
        Column('product_id', Integer, ForeignKey('tblWeHqppProduct.id'), primary_key=True),
        Column('tbs_id', Integer, ForeignKey('tblWeTbsLicense.license_id'), primary_key=True),
        Column('is_b2b', SmallInteger, nullable=False, server_default='0', primary_key=True),  # 0 = Not supplied, 1 = b2c, 2 = b2b
    )

    tables['tblWeCN8'] = tblWeCN8
    tables['tblWeUnuCode'] = tblWeUnuCode

    tables['tblWeTbsProductLevel'] = tblWeTbsProductLevel
    tables['tblWeTbsRegulationPeriodProduct'] = tblWeTbsRegulationPeriodProduct

    tables['tblWeCN8PeriodProduct'] = tblWeCN8PeriodProduct
    tables['tblWeHqppProduct'] = tblWeHqppProduct
    tables['tblWeHqppRegulationPeriodProduct'] = tblWeHqppRegulationPeriodProduct
    tables['tblWeHqppProductExcludedAtTbs'] = tblWeHqppProductExcludedAtTbs


    __tables_defined = True

def do_mappings():
    """
    This function is called by Django middleware when needed.
    """
    global __mapping_defined
    if __mapping_defined:
        return

    from basesoft.web.contrib.notification import models as notification_models
    from iqportal.weee.core import models as weee_models
    from we.product import models as product_models
    from we.tbs import models as tbs_models
    from we.member import models as member_models

    mapper(product_models.WeCN8, tables['tblWeCN8'],
        properties = {
            'unu_code': relation(product_models.WeUnuCode, backref='cn8_codes'),
            'weee_product_level': relation(weee_models.WEEEProductLevel)
        }
    )

    mapper(product_models.WeUnuCode, tables['tblWeUnuCode'],
    )

    mapper(product_models.WeTbsProductLevel, tables['tblWeTbsProductLevel'],
        properties = {
            'tbs': relation(tbs_models.WeTbsLicense),
            'sublevels': relation(
                product_models.WeTbsProductLevel, backref=backref(
                    'parent_level', remote_side=[tables['tblWeTbsProductLevel'].c.id]
                ), order_by=tables['tblWeTbsProductLevel'].c.position
            ),
        },
        order_by=[tables['tblWeTbsProductLevel'].c.parent_level_id, tables['tblWeTbsProductLevel'].c.position]
    )

    mapper(product_models.WeTbsRegulationPeriodProduct, tables['tblWeTbsRegulationPeriodProduct'],
        properties = {
            'tbs_product_level': relation(product_models.WeTbsProductLevel),
            'tbs_regulation_period': relation(tbs_models.WeTbsRegulationPeriod, backref="period_products"),
        },
        order_by=[tables['tblWeTbsRegulationPeriodProduct'].c.tbs_regulation_period_id,
                  tables['tblWeTbsRegulationPeriodProduct'].c.position]
    )

    mapper(product_models.WeCN8PeriodProduct, tables['tblWeCN8PeriodProduct'],
        properties = {
            'period_product': relation(product_models.WeTbsRegulationPeriodProduct),
        }
    )

    mapper(product_models.WeHqppProduct, tables['tblWeHqppProduct'],
        properties = {
            'cn8_code': relation(product_models.WeCN8),
            'hqpp': relation(member_models.WeHqppLicense)
        }
    )

    mapper(product_models.WeHqppRegulationPeriodProduct, tables['tblWeHqppRegulationPeriodProduct'],
        properties = {
            'product': relation(product_models.WeHqppProduct),
            'tbs_regulation_period_product': relation(product_models.WeTbsRegulationPeriodProduct),
            'tbs_regulation_period': relation(tbs_models.WeTbsRegulationPeriod),
            'tbs_products': relation(product_models.WeTbsProductLevel, secondary=tables['tblWeTbsRegulationPeriodProduct'])
        }
    )

    mapper(product_models.WeHqppProductExcludedAtTbs, tables['tblWeHqppProductExcludedAtTbs'],
        properties = {
            'product': relation(product_models.WeHqppProduct),
            'tbs': relation(tbs_models.WeTbsLicense),
        }
    )

    mapper(product_models.WeRegulationPeriodNotification,
           inherits=notification_models.Notification,
           polymorphic_identity='WeRegulationPeriodNotification'
    )

    __mapping_defined = True

def fill_data(conn):
    pass

def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'WEEE Europe product')

def app_permissions():
    from django.utils.translation import ugettext as _
    return [

        license_models.Permission(u'we_view_weee_category', u"""
            This gives the right to view WEEE categories
            """, _("View weee categories")),

        license_models.Permission(u'we_manage_cn8', u"""
            This gives the right to manage Weee cn8 codes
            """, _("Manage cn8 code")),

        license_models.Permission(u'we_manage_tbs_product_level', u"""
            This gives the right to manage TBS product levels
            """, _("Manage tbs product levels")),

        license_models.Permission(u'we_view_tbs_product_level', u"""
            This gives the right to view TBS product levels
            """, _("View tbs product levels")),

        license_models.Permission(u'we_manage_fee', u"""
            This gives the right to manage TBS product level fee
            """, _("Manage tbs product level fee")),

        license_models.Permission(u'we_view_products', u"""
            This gives the right to view LPP products
            """, _("View products")),

        license_models.Permission(u'we_create_products', u"""
            This gives the right to create LPP products
            """, _("Create products")),

        license_models.Permission(u'we_manage_cn8_tbs_product_level', u"""
            This gives the right to import file with cn8 and TBS local category suggestions.
            """, _("Manage cn8 to tbs local category suggestions")),

    ]

