# Copyright (c) 2016 BaseSoft B.V.
# See LICENSE for details.

from django.utils import simplejson
from basesoft.util.ordered_dict import OrderedDict
from iqportal.iqube.core.controllers import IQubeAggregator, MondrianMemberSerialize, ConfigurableWidgetController, DataSourceController

from we.core.controllers import AbstractWeController
from we.product.models import WeHqppProduct, WeTbsProductLevel


class WeReportController(AbstractWeController):

    def __serialize_member(self, x):
        return {'label': x['member_caption'], 'value': MondrianMemberSerialize(x)}

    def get_tbses(self, datasource, cube, config, values):
        tbses = self.get_tbs_licenses()
        return self.__get_dimension_config(tbses, 'TBS')

    def get_lpps(self, datasource, cube, config, values):
        lpps = self.get_local_producer_partners()
        return self.__get_dimension_config(lpps, 'LPP')

    def get_producers(self, datasource, cube, config, values):
        producers = self.get_member_licenses()  # HQPP licenses
        return self.__get_dimension_config(producers, 'Producer')

    def get_producer_products(self, datasource, cube, config, values):
        cwc = ConfigurableWidgetController(self.sa_session, account_id=self.account_id)
        dsc = DataSourceController(self.sa_session, account_id=self.account_id)

        ds = dsc.get_datasource(datasource)
        options = cwc.get_dimension_options(ds, cube, config, values)
        return self.__get_product_dimension_config(options)

    def get_local_categories(self, datasource, cube, config, values):
        cwc = ConfigurableWidgetController(self.sa_session, account_id=self.account_id)
        dsc = DataSourceController(self.sa_session, account_id=self.account_id)

        ds = dsc.get_datasource(datasource)
        options = cwc.get_dimension_options(ds, cube, config, values)
        return self.__get_local_category_dimension_config(options)

    def __get_dimension_config(self, members, dimension):

        rv = OrderedDict()
        iqube_members = []
        all_members = [
            self.__serialize_member({
                "dimension_unique_name": "[%s]" % dimension,
                "depth": None,
                "member_unique_name": "[%s].[All %ss]" % (dimension, dimension),
                "member_caption": "All %ss" % dimension
            }),
            self.__serialize_member({
                "hierarchy_unique_name": "[%s]" % dimension,
                "dimension_unique_name": "[%s]" % dimension,
                "level_unique_name": "[%s].[%s]" % (dimension, dimension),
                "member_unique_name": "[%s].[%s].Members" % (dimension, dimension),
                "member_name": "[%s].[%s]" % (dimension, dimension),
                "member_caption": "%s (Members)" % dimension
            })
        ]

        for member in members:
            caption = member.name
            if dimension == 'TBS':
                caption = "%s - %s" % (member.name, member.company.addresses[0].address.country.name)
            iqube_member = {
                "member_caption": caption,
                "dimension_unique_name": "[%s]" % dimension,
                "member_unique_name": "[%s].[%s]" % (dimension, member.id),
                "depth": "1",
                "level_unique_name": "[%s].[%s]" % (dimension, dimension),
                "member_name": "%s" % member.id,
                "hierarchy_unique_name": "[%s]" % dimension
            }

            iqube_members.append(iqube_member)

        iqube_members = [self.__serialize_member(iqube_member) for iqube_member in sorted(iqube_members, key=lambda k: k['member_caption'].lower())]
        rv.update([('All %ss' % dimension, all_members, )])
        rv.update([('%ss' % dimension, iqube_members, )])

        return rv

    def __get_product_dimension_config(self, members):

        for member in members['Product']:
            m = simplejson.loads(member['value'])
            sku = self.sa_session.query(WeHqppProduct).filter(WeHqppProduct.id == m['member_name']).one()
            member['label'] = "%s - %s" % (sku.hqpp.name, sku.sku)

        members['Product'] = sorted(members['Product'], key=lambda k: k['label'])

        return members

    def __get_local_category_dimension_config(self, members):
        for member in members['Local Category']:
            m = simplejson.loads(member['value'])
            lc = self.sa_session.query(WeTbsProductLevel).filter(WeTbsProductLevel.id == m['member_name']).one()
            member['label'] = "%s - %s" % (lc.tbs.name, lc.label)

        members['Local Category'] = sorted(members['Local Category'], key=lambda k: k['label'])

        return members


class WeDeclarationAggregator(IQubeAggregator):
    _table = "aggWeDeclaration"
    _affected_cubes = ({'datasource': 'WeDeclarationData', 'cube': 'WeDeclarationDataCube'}, )
    _columns = [
        'declaration_id',
        'declaration_line_id',
        'declaration_status',
        'lpp_id',
        'country_id',
        'tbs_id',
        'hqpp_id',
        'year',
        'month',
        'hqpp_period_product_id',
        'tbs_product_level_id',
        'weee_product_level_id',
        'cn8',
        'unu_code_id',
        'quantity',
        'weight',
        'amount_excluding_vat',
        'last_updated'
    ]

    _select_sql = """

        SELECT
            d.id as declaration_id,
            dl.id as declaration_line_id,
            d.status as declaration_status,
            lpp.id as lpp_id,
            c.id as country_id,
            lpp.tbs_id as tbs_id,
            lpp.hqpp_id as hqpp_id,
            d.year as `year`,
            d.month as `month`,
            hrpp.id as hqpp_period_product_id,
            trpp.tbs_product_level_id as tbs_product_level_id,
            cn8.weee_product_level_id as weee_product_level_id,
            hp.cn8 as cn8,
            cn8.unu_code_id as unu_code_id,
            dl.quantity as quantity,
            dl.product_total_weight as weight,
            dl.product_total_fee as amount_excluding_vat,
            d.last_updated as last_updated
        FROM tblWeDeclarationLine as dl
        INNER JOIN tblWeDeclaration as d
            ON dl.declaration_id = d.id
        INNER JOIN tblWeLpp as lpp
            ON d.lpp_id = lpp.id
        INNER JOIN tblWeTbsLicense as tl
            ON lpp.tbs_id = tl.license_id
        INNER JOIN tblWeHqppLicense as hl
            ON lpp.hqpp_id = hl.license_id

        INNER JOIN tblWeHqppProduct as hp
            ON hp.hqpp_id = hl.license_id
        INNER JOIN tblWeHqppRegulationPeriodProduct as hrpp
            ON hp.id = hrpp.product_id  AND dl.hqpp_period_product_id = hrpp.id
        INNER JOIN tblWeTbsRegulationPeriodProduct as trpp
            ON hrpp.tbs_regulation_period_product_id = trpp.id
        INNER JOIN tblWeCN8 as cn8
            ON hp.cn8 = cn8.code
        INNER JOIN tblLicense as l
            ON hl.license_id = l.id
        INNER JOIN tblCompany as comp
            ON l.company_id = comp.id
        INNER JOIN tblCompanyAddress as compad
            ON comp.id = compad.company_id
        INNER JOIN tblAddress as ad
            ON compad.address_id = ad.id
        INNER JOIN tblCountry as c
            ON ad.country_id = c.id
        WHERE d.last_updated > %(max_last_updated)s
        GROUP BY dl.id
        ORDER BY d.id
    """

    def run(self, loop_validator=None):

        affected_rows = 0

        max_last_updated = 0
        rows = self.db_engine_ro.execute("""
            SELECT IFNULL(MAX(a.last_updated), 0)
            FROM %s as a""" % self._table)

        row = rows.fetchone()
        rows.close()

        if row is not None:
            max_last_updated = row[0]

        self.db_engine.execute("""
            DELETE FROM %(tbl)s WHERE declaration_id IN (
                SELECT id
                FROM tblWeDeclaration
                WHERE last_updated > '%(max_last_updated)s'
            )""" % {
            'tbl': self._table,
            'max_last_updated': max_last_updated
        })

        affected_rows = affected_rows + self._work({
            'max_last_updated': max_last_updated,
        }, loop_validator)

        return affected_rows


class WeOrderAggregator(IQubeAggregator):
    _table = "aggWeOrder"
    _affected_cubes = ({'datasource': 'WeDeclarationData', 'cube': 'WeOrderCube'}, )

    _columns = [
        'order_id',
        'order_line_id',
        'year',
        'biannual',
        'quarter',
        'month',
        'tbs_id',
        'hqpp_id',
        'lpp_id',
        'country_id',
        'tbs_product_level_id',
        'is_b2b',
        'quantity',
        'weight',
        'line_amount_excluding_vat',
        'last_updated'
    ]

    _select_sql = """
    SELECT
            do.order_id as order_id,
            dol.order_line_id as order_line_id,
            do.year as `year`,
            IF(ISNULL(do.bi_annual) AND NOT ISNULL(do.month), IF(do.month <= 6, 1, 2), do.bi_annual) as `biannual`,
            IF(ISNULL(do.quarter) AND NOT ISNULL(do.month), IF(do.month < 4, 1, IF(do.month < 7, 2, IF(do.month < 10, 3, 4))), do.quarter) as `quarter`,
            do.month as `month`,
            o.merchant_id as tbs_id,
            o.client_id as hqpp_id,
            do.lpp_id as lpp_id,
            c.id as country_id,
            IF(ISNULL(o.display_date), NULL, DATE_FORMAT(o.display_date, '%%Y%%m%%d')) as submitted_time_id,
            dol.tbs_product_level_id as tbs_product_level_id,
            dol.is_b2b as is_b2b,
            ol.quantity as quantity,
            dol.weight as weight,
            ol.line_amount_excluding_vat as line_amount_excluding_vat,
            o.last_updated as last_updated
        FROM tblOrder as o
        INNER JOIN tblOrderLine as ol
            ON o.id = ol.order_id
        INNER JOIN tblWeDeclarationOrder as do
            ON o.id = do.order_id
        INNER JOIN tblWeDeclarationOrderLine as dol
            ON ol.id = dol.order_line_id
        INNER JOIN tblWeHqppLicense as hl
            ON o.client_id = hl.license_id
        INNER JOIN tblLicense as l
            ON hl.license_id = l.id
        INNER JOIN tblCompany as comp
            ON l.company_id = comp.id
        INNER JOIN tblCompanyAddress as compad
            ON comp.id = compad.company_id
        INNER JOIN tblAddress as ad
            ON compad.address_id = ad.id
        INNER JOIN tblCountry as c
            ON ad.country_id = c.id
        WHERE o.last_updated > %(max_last_updated)s
        AND o.deleted = 0
        ORDER BY o.id

    """

    def run(self, loop_validator=None):

        affected_rows = 0

        max_last_updated = 0
        rows = self.db_engine_ro.execute("""
            SELECT IFNULL(MAX(a.last_updated), 0)
            FROM %s as a""" % self._table)

        row = rows.fetchone()
        rows.close()

        if row is not None:
            max_last_updated = row[0]

        self.db_engine.execute("""
            DELETE FROM %(tbl)s WHERE order_id IN (
                SELECT id
                FROM tblOrder
                WHERE last_updated > '%(max_last_updated)s'
            )""" % {
            'tbl': self._table,
            'max_last_updated': max_last_updated
        })

        affected_rows = affected_rows + self._work({
            'max_last_updated': max_last_updated,
        }, loop_validator)

        return affected_rows