# Copyright (c) 2016 BaseSoft B.V.
# See LICENSE for details.

from django.conf import settings
from django.utils.translation import ugettext as _

from basesoft.logging import getLogger
from basesoft.web.core.exceptions import PublicSecurityException
from basesoft.web.core.management.commands.iqubesync import CubeBase, ProjectBase, RoleBase
from basesoft.web.core.util.controllers import AbstractController

from iqportal.iqube.core.controllers import IQubeRole, IQubeRoleController
from iqportal.iqube.core.models import DataSource
from iqportal.weee.core.models import WEEEProductLevel

from we.core.controllers import AbstractWeController
from we.core.models import WeOperatorLicense
from we.member.models import WeHqppLicense
from we.product.controllers import WeProductController
from we.tbs.models import WeTbsLicense

logger = getLogger('we.report.iqube_sync')


class WeDeclarationDataProject(ProjectBase):
    project_name = 'WeDeclarationData'
    world_privileges = DataSource.PRIV_VIEW
    description = 'WEEE-EU declaration data'
    role_level = DataSource.ROLE_LEVEL_DB

    def __init__(self, sa_session, controller):
        ProjectBase.__init__(self, sa_session, controller)
        self._db_from_connstr(settings.DB_CONNSTR_READONLY)

    def get_cubes(self):
        return WeDeclarationDataCube, WeOrderCube,

    def get_roles(self):
        return WeDeclarationDataCubeRole, WeOrderCubeRole,


class WeDeclarationDataCube(CubeBase):
    name = 'WeDeclarationDataCube'

    def _xml(self):
        return """
            <Cube name="%(name)s">

                <Table name="aggWeDeclaration" />

                <Dimension name="LPP" foreignKey="lpp_id">
                    <Hierarchy hasAll="true" allMemberName="All LPPs" primaryKey="id" primaryKeyTable="tblWeLpp">
                            <Table name="tblWeLpp" />
                            <Level name="LPP" table="tblWeLpp" column="id" captionColumn="name" type="Integer" ordinalColumn="name" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="TBS" foreignKey="tbs_id">
                    <Hierarchy hasAll="true" allMemberName="All TBSs" primaryKey="license_id" primaryKeyTable="tblWeTbsLicense">
                    <Join leftKey="license_id" rightKey="id">
                        <Table name="tblWeTbsLicense"/>
                        <Table name="tblLicense" alias="tbs" />
                    </Join>
                        <Level name="TBS" table="tbs" column="id" captionColumn="name" type="Integer" ordinalColumn="name" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="Producer" foreignKey="hqpp_id">
                    <Hierarchy hasAll="true" allMemberName="All Producers" primaryKey="license_id" primaryKeyTable="tblWeHqppLicense">
                    <Join leftKey="license_id" rightKey="id">
                        <Table name="tblWeHqppLicense"/>
                        <Table name="tblLicense" alias="hqpp" />
                    </Join>
                        <Level name="Producer" table="hqpp" column="id" captionColumn="name" type="Integer" ordinalColumn="name" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="Country" foreignKey="country_id">
                    <Hierarchy hasAll="true" allMemberName="All Countries" primaryKey="id" primaryKeyTable="tblCountry">
                        <Table name="tblCountry" />
                        <Level name="Country" table="tblCountry" column="id" captionColumn="name" type="Integer" ordinalColumn="name"/>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Declaration Year" foreignKey="declaration_id">
                    <Hierarchy hasAll="true" allMemberName="All Years" primaryKey="id" primaryKeyTable="tblWeDeclaration">
                        <Table name="tblWeDeclaration"/>
                        <Level name="Year" column="year" ordinalColumn="year" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="Declaration Month" foreignKey="declaration_id">
                    <Hierarchy hasAll="true" allMemberName="All Months" primaryKey="id" primaryKeyTable="tblWeDeclaration">
                        <Table name="tblWeDeclaration"/>
                        <Level name="Month" column="month" ordinalColumn="month">
                            <CaptionExpression>
                                <SQL dialect="mysql">
                                    MONTHNAME(STR_TO_DATE(`tblWeDeclaration`.`month`, '%%m'))
                                </SQL>
                            </CaptionExpression>
                        </Level>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Declaration Status" foreignKey="declaration_id">
                    <Hierarchy hasAll="true" allMemberName="All Declaration Statuses" primaryKey="id" primaryKeyTable="tblWeDeclaration">
                        <Table name="tblWeDeclaration"/>
                        <Level name="Declaration Status" column="status" ordinalColumn="status" >
                            <CaptionExpression>
                                <SQL dialect="mysql">
                                    CASE `tblWeDeclaration`.`status`
                                        WHEN 1 THEN 'Complete'
                                        WHEN 2 THEN 'Exported'
                                        WHEN 3 THEN 'Resubmit'
                                        ELSE '-' END
                                </SQL>
                            </CaptionExpression>
                        </Level>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Product" foreignKey="hqpp_period_product_id">
                    <Hierarchy hasAll="true" allMemberName="All Products" primaryKey="id" primaryKeyTable="tblWeHqppRegulationPeriodProduct">
                        <Join leftKey="product_id" rightKey="id" >
                            <Table name="tblWeHqppRegulationPeriodProduct" />
                            <Table name="tblWeHqppProduct" />
                        </Join>
                            <Level name="Product" table="tblWeHqppProduct" column="id" captionColumn="sku" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="Local Category" foreignKey="tbs_product_level_id">
                    <Hierarchy hasAll="true" allMemberName="All Local categories" primaryKey="id" primaryKeyTable="tblWeTbsProductLevel">
                        <Table name="tblWeTbsProductLevel" />
                        <Level name="Local Category" table="tblWeTbsProductLevel" column="id" captionColumn="label" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="CN8" foreignKey="cn8">
                    <Hierarchy hasAll="true" allMemberName="All CN8s" primaryKey="code" primaryKeyTable="tblWeCN8">
                        <Table name="tblWeCN8" />
                        <Level name="CN8" table="tblWeCN8" column="code" captionColumn="code" ordinalColumn="code" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="WEEE Category" foreignKey="weee_product_level_id">
                    <Hierarchy hasAll="true" primaryKey="id" allMemberName="All WEEE Categories" primaryKeyTable="tblWEEEProductLevel">
                        <Table name="tblWEEEProductLevel" />
                        <Level name="WEEE Category" table="tblWEEEProductLevel" column="id" captionColumn="label" type="Integer" ordinalColumn="position"
                            uniqueMembers="true">
                        </Level>
                    </Hierarchy>
                </Dimension>

                <Measure name="Quantity" column="quantity" aggregator="sum" datatype="Integer" />
                <Measure name="Weight" column="weight" aggregator="sum" datatype="Numeric" formatString="#,##0.000" />
                <Measure name="Amount" column="amount_excluding_vat" aggregator="sum" datatype="Numeric" formatString="#,##0.00" />
                <Measure name="Number of declarations" column="declaration_id" aggregator="distinct-count" datatype="Integer" />

            </Cube>
                """ % {'name': self.name}

    xml = property(_xml)


class WeDeclarationDataCubeRole(RoleBase):
    CUBE = 'WeDeclarationDataCube'

    def get_role_list(self, account_id):

        logger.info("get_role_list: %s" % account_id)

        # check required permissions
        ac = AbstractController(self.sa_session, account_id=account_id, check_active=False)
        ac.assert_permissions('we_view_report')

        current_license = ac.get_default_group().license

        rc = IQubeRoleController(self.sa_session, account_id, self.ds, self.CUBE)
        rc.grant_cube(IQubeRole.ACCESS_ALL)

        c = AbstractWeController(self.sa_session, account_id=account_id)
        pc = WeProductController(self.sa_session, account_id=account_id)
        lc = False

        categories = c.get_privileged_object(WEEEProductLevel, priv=WEEEProductLevel.PRIV_VIEW)

        for cat in categories:
           rc.grant_member('[WEEE Category].[%d]' % cat.id, IQubeRole.ACCESS_ALL)
        rc.grant_hierarchy('[WEEE Category]', IQubeRole.ACCESS_CUSTOM, IQubeRole.ROLLUP_POLICY_PARTIAL)

        if isinstance(current_license, WeOperatorLicense):
            return rc.to_list()

        elif isinstance(current_license, WeTbsLicense):
            tbss = []
            for tbs in c.get_tbs_licenses():
                if tbs.id not in tbss:
                    for lp in pc.get_tbs_product_levels(None, tbs_id=tbs.id):
                        lc = True
                        rc.grant_member('[Local Category].[%d]' % lp.id, IQubeRole.ACCESS_ALL)
                    rc.grant_member('[TBS].[%s]' % tbs.license_id, IQubeRole.ACCESS_ALL)
                    tbss.append(tbs.license_id)
                    rc.grant_hierarchy('[Producer]', IQubeRole.ACCESS_NONE, IQubeRole.ROLLUP_POLICY_PARTIAL)

        elif isinstance(current_license, WeHqppLicense):
            tbss = []
            members = []

            hqpps = c.get_member_licenses()
            for hqpp in hqpps:
                lpps = c.get_local_producer_partners(hqpp.id)

                for lpp in lpps:
                    if lpp.tbs_id not in tbss:
                        for lp in pc.get_tbs_product_levels(None, tbs_id=lpp.tbs_id):
                            lc = True
                            rc.grant_member('[Local Category].[%d]' % lp.id, IQubeRole.ACCESS_ALL)
                        rc.grant_member('[TBS].[%s]' % lpp.tbs_id, IQubeRole.ACCESS_ALL)
                        tbss.append(lpp.tbs.id)
                    if lpp.hqpp_id not in members:
                        rc.grant_member('[Producer].[%s]' % lpp.hqpp_id, IQubeRole.ACCESS_ALL)
                        members.append(lpp.hqpp_id)
            rc.grant_hierarchy('[Producer]', IQubeRole.ACCESS_CUSTOM, IQubeRole.ROLLUP_POLICY_PARTIAL)

        else:
            raise PublicSecurityException(_('Invalid license type'))

        rc.grant_hierarchy('[TBS]', IQubeRole.ACCESS_CUSTOM, IQubeRole.ROLLUP_POLICY_PARTIAL)
        if lc:
            rc.grant_hierarchy('[Local Category]', IQubeRole.ACCESS_CUSTOM, IQubeRole.ROLLUP_POLICY_PARTIAL)
        else:
            rc.grant_hierarchy('[Local Category]', IQubeRole.ACCESS_NONE, IQubeRole.ROLLUP_POLICY_PARTIAL)

        return rc.to_list()


class WeOrderCube(CubeBase):
    name = 'WeOrderCube'

    def _xml(self):
        return """
            <Cube name="%(name)s" defaultMeasure="Amount (excl. vat)">

                <Table name="aggWeOrder" />

                <Dimension name="Order" foreignKey="order_id">
                    <Hierarchy hasAll="true" allMemberName="All Orders" primaryKey="id" primaryKeyTable="tblWeDeclarationOrder">
                            <Table name="tblWeDeclarationOrder"/>
                        <Level name="Order" table="tblWeDeclarationOrder" column="id" type="Integer" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="Order Line" foreignKey="order_line_id">
                    <Hierarchy hasAll="true" allMemberName="All Order Lines" primaryKey="id" primaryKeyTable="tblWeDeclarationOrderLine">
                            <Table name="tblWeDeclarationOrderLine"/>
                        <Level name="Order Line" table="tblWeDeclarationOrderLine" column="id" type="Integer" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="TBS" foreignKey="tbs_id">
                    <Hierarchy hasAll="true" allMemberName="All TBSs" primaryKey="license_id" primaryKeyTable="tblWeTbsLicense">
                        <Join leftKey="license_id" rightKey="id">
                            <Table name="tblWeTbsLicense"/>
                            <Table name="tblLicense" alias="tbs" />
                        </Join>
                        <Level name="TBS" table="tbs" column="id" captionColumn="name" type="Integer" ordinalColumn="name" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="Producer" foreignKey="hqpp_id">
                    <Hierarchy hasAll="true" allMemberName="All Producers" primaryKey="license_id" primaryKeyTable="tblWeHqppLicense">
                        <Join leftKey="license_id" rightKey="id">
                            <Table name="tblWeHqppLicense"/>
                            <Table name="tblLicense" alias="hqpp" />
                        </Join>
                        <Level name="Producer" table="hqpp" column="id" captionColumn="name" type="Integer" ordinalColumn="name"/>
                    </Hierarchy>
                </Dimension>

                <Dimension name="LPP" foreignKey="lpp_id">
                    <Hierarchy hasAll="true" allMemberName="All LPPs" primaryKey="id" primaryKeyTable="tblWeLpp">
                            <Table name="tblWeLpp" />
                            <Level name="LPP" table="tblWeLpp" column="id" captionColumn="name" type="Integer" ordinalColumn="name" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="Country" foreignKey="country_id">
                    <Hierarchy hasAll="true" allMemberName="All Countries" primaryKey="id" primaryKeyTable="tblCountry">
                        <Table name="tblCountry" />
                        <Level name="Country" table="tblCountry" column="id" captionColumn="name" type="Integer" ordinalColumn="name"/>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Order Year">
                    <Hierarchy hasAll="true" allMemberName="All Declaration Years">
                        <Level name="Year" column="year"/>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Order Bi-annual">
                    <Hierarchy hasAll="true" allMemberName="All Order Bi-annual">
                        <Level name="Bi-annual" column="biannual"/>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Order Quarter">
                    <Hierarchy hasAll="true" allMemberName="All Declaration Quarters">
                        <Level name="Quarter" column="quarter">
                            <CaptionExpression>
                                <SQL dialect="mysql">
                                    CONCAT('Q', `aggWeOrder`.`quarter`)
                                </SQL>
                            </CaptionExpression>
                        </Level>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Order Month">
                    <Hierarchy hasAll="true" allMemberName="All Declaration Months">
                        <Level name="Month" column="month">
                            <CaptionExpression>
                                <SQL dialect="mysql">
                                    MONTHNAME(STR_TO_DATE(`aggWeOrder`.`month`, '%%m'))
                                </SQL>
                            </CaptionExpression>
                        </Level>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Local Category" foreignKey="tbs_product_level_id">
                    <Hierarchy hasAll="true" allMemberName="All Local categories" primaryKey="id" primaryKeyTable="tblWeTbsProductLevel">
                        <Table name="tblWeTbsProductLevel" />
                        <Level name="Local Category" table="tblWeTbsProductLevel" column="id" captionColumn="label" />
                    </Hierarchy>
                </Dimension>


                <Dimension name="Is b2b" foreignKey="is_b2b">
                    <Hierarchy hasAll="true" allMemberName="All b2b/b2c">
                            <Level name="Is b2b" column="is_b2b">
                                <CaptionExpression>
                                    <SQL dialect="mysql">
                                        CASE `aggWeOrder`.`is_b2b`
                                            WHEN 0 THEN 'N/A'
                                            WHEN 1 THEN 'B2C'
                                            WHEN 2 THEN 'B2B'
                                            ELSE '-' END
                                    </SQL>
                                </CaptionExpression>
                            </Level>
                    </Hierarchy>
                </Dimension>

                <Measure name="Weight" column="weight" aggregator="sum" datatype="Numeric" formatString="#,##0.000" />
                <Measure name="Quantity" column="quantity" aggregator="sum" datatype="Integer" />
                <Measure name="Average quantity" column="quantity" aggregator="avg" datatype="Integer" />
                <Measure name="Amount (excl. vat)" column="line_amount_excluding_vat" aggregator="sum" datatype="Numeric" formatString="#,##0.00" />

            </Cube>
        """ % {'name': self.name}
    xml = property(_xml)


class WeOrderCubeRole(RoleBase):
    CUBE = 'WeOrderCube'

    def get_role_list(self, account_id):

        logger.info("get_role_list: %s" % account_id)

        # check required permissions
        ac = AbstractController(self.sa_session, account_id=account_id, check_active=False)
        ac.assert_permissions('we_view_report')

        current_license = ac.get_default_group().license

        rc = IQubeRoleController(self.sa_session, account_id, self.ds, self.CUBE)
        rc.grant_cube(IQubeRole.ACCESS_ALL)

        c = AbstractWeController(self.sa_session, account_id=account_id)

        if isinstance(current_license, WeOperatorLicense):
            return rc.to_list()

        elif isinstance(current_license, WeTbsLicense):
            tbss = []
            for tbs in c.get_tbs_licenses():
                if tbs.id not in tbss:
                    rc.grant_member('[TBS].[%s]' % tbs.license_id, IQubeRole.ACCESS_ALL)
                    tbss.append(tbs.license_id)
                    rc.grant_hierarchy('[Producer]', IQubeRole.ACCESS_NONE, IQubeRole.ROLLUP_POLICY_PARTIAL)

        elif isinstance(current_license, WeHqppLicense):
            tbss = []
            members = []

            hqpps = c.get_member_licenses()

            for hqpp in hqpps:
                lpps = c.get_local_producer_partners(hqpp.id)

                for lpp in lpps:
                    if lpp.tbs_id not in tbss:
                        rc.grant_member('[TBS].[%s]' % lpp.tbs_id, IQubeRole.ACCESS_ALL)
                        tbss.append(lpp.tbs.id)
                    if lpp.hqpp_id not in members:
                        rc.grant_member('[Producer].[%s]' % lpp.hqpp_id, IQubeRole.ACCESS_ALL)
                        members.append(lpp.hqpp_id)
            rc.grant_hierarchy('[Producer]', IQubeRole.ACCESS_CUSTOM, IQubeRole.ROLLUP_POLICY_PARTIAL)

        else:
            raise PublicSecurityException(_('Invalid license type'))

        rc.grant_hierarchy('[TBS]', IQubeRole.ACCESS_CUSTOM, IQubeRole.ROLLUP_POLICY_PARTIAL)

        return rc.to_list()


projects = (WeDeclarationDataProject, )